import click
import json as pyjson

from toot import api
from toot.cli import cli, pass_context, json_option, Context
from toot.entities import Tag, from_dict
from toot.output import print_tag_list, print_warning


@cli.group()
def tags():
    """List, follow, and unfollow tags"""


@tags.command()
@click.argument("tag")
@json_option
@pass_context
def info(ctx: Context, tag, json: bool):
    """Show a hashtag and its associated information"""
    tag = api.find_tag(ctx.app, ctx.user, tag)

    if not tag:
        raise click.ClickException("Tag not found")

    if json:
        click.echo(pyjson.dumps(tag))
    else:
        tag = from_dict(Tag, tag)
        click.secho(f"#{tag.name}", fg="yellow")
        click.secho(tag.url, italic=True)
        if tag.following:
            click.echo("Followed")
        else:
            click.echo("Not followed")


@tags.command()
@json_option
@pass_context
def followed(ctx: Context, json: bool):
    """List followed tags"""
    tags = api.followed_tags(ctx.app, ctx.user)
    if json:
        click.echo(pyjson.dumps(tags))
    else:
        if tags:
            print_tag_list(tags)
        else:
            click.echo("You're not following any hashtags")


@tags.command()
@click.argument("tag")
@json_option
@pass_context
def follow(ctx: Context, tag: str, json: bool):
    """Follow a hashtag"""
    tag = tag.lstrip("#")
    response = api.follow_tag(ctx.app, ctx.user, tag)
    if json:
        click.echo(response.text)
    else:
        click.secho(f"✓ You are now following #{tag}", fg="green")


@tags.command()
@click.argument("tag")
@json_option
@pass_context
def unfollow(ctx: Context, tag: str, json: bool):
    """Unfollow a hashtag"""
    tag = tag.lstrip("#")
    response = api.unfollow_tag(ctx.app, ctx.user, tag)
    if json:
        click.echo(response.text)
    else:
        click.secho(f"✓ You are no longer following #{tag}", fg="green")


@tags.command()
@json_option
@pass_context
def featured(ctx: Context, json: bool):
    """List hashtags featured on your profile."""
    response = api.featured_tags(ctx.app, ctx.user)
    if json:
        click.echo(response.text)
    else:
        tags = response.json()
        if tags:
            print_tag_list(tags)
        else:
            click.echo("You don't have any featured hashtags")


@tags.command()
@click.argument("tag")
@json_option
@pass_context
def feature(ctx: Context, tag: str, json: bool):
    """Feature a hashtag on your profile"""
    tag = tag.lstrip("#")
    response = api.feature_tag(ctx.app, ctx.user, tag)
    if json:
        click.echo(response.text)
    else:
        click.secho(f"✓ Tag #{tag} is now featured", fg="green")


@tags.command()
@click.argument("tag")
@json_option
@pass_context
def unfeature(ctx: Context, tag: str, json: bool):
    """Unfollow a hashtag

    TAG can either be a tag name like "#foo" or "foo" or a tag ID.
    """
    featured_tag = api.find_featured_tag(ctx.app, ctx.user, tag)

    # TODO: should this be idempotent?
    if not featured_tag:
        raise click.ClickException(f"Tag {tag} is not featured")

    response = api.unfeature_tag(ctx.app, ctx.user, featured_tag["id"])
    if json:
        click.echo(response.text)
    else:
        click.secho(f"✓ Tag #{featured_tag['name']} is no longer featured", fg="green")


# -- Deprecated commands -------------------------------------------------------

@cli.command(name="tags_followed", hidden=True)
@pass_context
def tags_followed(ctx: Context):
    """List hashtags you follow"""
    print_warning("`toot tags_followed` is deprecated in favour of `toot tags followed`")
    response = api.followed_tags(ctx.app, ctx.user)
    print_tag_list(response)


@cli.command(name="tags_follow", hidden=True)
@click.argument("tag")
@pass_context
def tags_follow(ctx: Context, tag: str):
    """Follow a hashtag"""
    print_warning("`toot tags_follow` is deprecated in favour of `toot tags follow`")
    tag = tag.lstrip("#")
    api.follow_tag(ctx.app, ctx.user, tag)
    click.secho(f"✓ You are now following #{tag}", fg="green")


@cli.command(name="tags_unfollow", hidden=True)
@click.argument("tag")
@pass_context
def tags_unfollow(ctx: Context, tag: str):
    """Unfollow a hashtag"""
    print_warning("`toot tags_unfollow` is deprecated in favour of `toot tags unfollow`")
    tag = tag.lstrip("#")
    api.unfollow_tag(ctx.app, ctx.user, tag)
    click.secho(f"✓ You are no longer following #{tag}", fg="green")



"""
Helpers for testing.
"""

import time
from typing import Any, Callable


class MockResponse:
    def __init__(self, response_data={}, ok=True, is_redirect=False):
        self.response_data = response_data
        self.content = response_data
        self.ok = ok
        self.is_redirect = is_redirect

    def raise_for_status(self):
        pass

    def json(self):
        return self.response_data


def retval(val):
    return lambda *args, **kwargs: val


def run_with_retries(fn: Callable[..., Any]):
    """
    Run the the given function repeatedly until it finishes without raising an
    AssertionError. Sleep a bit between attempts. If the function doesn't
    succeed in the given number of tries raises the AssertionError. Used for
    tests which should eventually succeed.
    """

    # Wait upto 6 seconds with incrementally longer sleeps
    delays = [0.1, 0.2, 0.3, 0.4, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]

    for delay in delays:
        try:
            return fn()
        except AssertionError:
            time.sleep(delay)

    fn()



from requests import Request, Session
from requests.exceptions import RequestException

from toot import __version__
from toot.exceptions import NotFoundError, ApiError
from toot.logging import log_request, log_request_exception, log_response


def send_request(request, allow_redirects=True):
    # Set a user agent string
    # Required for accessing instances using Cloudfront DDOS protection.
    request.headers["User-Agent"] = "toot/{}".format(__version__)

    log_request(request)

    try:
        with Session() as session:
            prepared = session.prepare_request(request)
            settings = session.merge_environment_settings(prepared.url, {}, None, None, None)
            response = session.send(prepared, allow_redirects=allow_redirects, **settings)
    except RequestException as ex:
        log_request_exception(request, ex)
        raise ApiError(f"Request failed: {str(ex)}")

    log_response(response)

    return response


def _get_error_message(response):
    """Attempt to extract an error message from response body"""
    try:
        data = response.json()
        if "error_description" in data:
            return data['error_description']
        if "error" in data:
            return data['error']
    except Exception:
        pass

    return f"Unknown error: {response.status_code} {response.reason}"


def process_response(response):
    if not response.ok:
        error = _get_error_message(response)

        if response.status_code == 404:
            raise NotFoundError(error)

        raise ApiError(error)

    return response


def get(app, user, path, params=None, headers=None):
    url = app.base_url + path

    headers = headers or {}
    headers["Authorization"] = f"Bearer {user.access_token}"

    request = Request('GET', url, headers, params=params)
    response = send_request(request)

    return process_response(response)


def anon_get(url, params=None):
    request = Request('GET', url, None, params=params)
    response = send_request(request)

    return process_response(response)


def post(app, user, path, headers=None, files=None, data=None, json=None, allow_redirects=True):
    url = app.base_url + path

    headers = headers or {}
    headers["Authorization"] = f"Bearer {user.access_token}"

    return anon_post(url, headers=headers, files=files, data=data, json=json, allow_redirects=allow_redirects)


def anon_put(url, headers=None, files=None, data=None, json=None, allow_redirects=True):
    request = Request(method="PUT", url=url, headers=headers, files=files, data=data, json=json)
    response = send_request(request, allow_redirects)

    return process_response(response)


def put(app, user, path, headers=None, files=None, data=None, json=None, allow_redirects=True):
    url = app.base_url + path

    headers = headers or {}
    headers["Authorization"] = f"Bearer {user.access_token}"

    return anon_put(url, headers=headers, files=files, data=data, json=json, allow_redirects=allow_redirects)


def patch(app, user, path, headers=None, files=None, data=None, json=None):
    url = app.base_url + path

    headers = headers or {}
    headers["Authorization"] = f"Bearer {user.access_token}"

    request = Request('PATCH', url, headers=headers, files=files, data=data, json=json)
    response = send_request(request)

    return process_response(response)


def delete(app, user, path, data=None, json=None, headers=None):
    url = app.base_url + path

    headers = headers or {}
    headers["Authorization"] = f"Bearer {user.access_token}"

    request = Request('DELETE', url, headers=headers, data=data, json=json)
    response = send_request(request)

    return process_response(response)


def anon_post(url, headers=None, files=None, data=None, json=None, allow_redirects=True):
    request = Request(method="POST", url=url, headers=headers, files=files, data=data, json=json)
    response = send_request(request, allow_redirects)

    return process_response(response)



"""
Dataclasses which represent entities returned by the Mastodon API.

Data classes my have an optional static method named `__toot_prepare__` which is
used when constructing the data class using `from_dict`. The method will be
called with the dict and may modify it and return a modified dict. This is used
to implement any pre-processing which may be required, e.g. to support
different versions of the Mastodon API.
"""

import dataclasses

from dataclasses import dataclass, is_dataclass
from datetime import date, datetime
from functools import lru_cache
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union
from typing import get_type_hints

from toot.typing_compat import get_args, get_origin
from toot.utils import get_text
from toot.utils.datetime import parse_datetime


@dataclass
class AccountField:
    """
    https://docs.joinmastodon.org/entities/Account/#Field
    """
    name: str
    value: str
    verified_at: Optional[datetime]


@dataclass
class CustomEmoji:
    """
    https://docs.joinmastodon.org/entities/CustomEmoji/
    """
    shortcode: str
    url: str
    static_url: str
    visible_in_picker: bool
    category: str


@dataclass
class Account:
    """
    https://docs.joinmastodon.org/entities/Account/
    """
    id: str
    username: str
    acct: str
    url: str
    display_name: str
    note: str
    avatar: str
    avatar_static: str
    header: str
    header_static: str
    locked: bool
    fields: List[AccountField]
    emojis: List[CustomEmoji]
    bot: bool
    group: bool
    discoverable: Optional[bool]
    noindex: Optional[bool]
    moved: Optional["Account"]
    suspended: Optional[bool]
    limited: Optional[bool]
    created_at: datetime
    last_status_at: Optional[date]
    statuses_count: int
    followers_count: int
    following_count: int
    source: Optional[dict]

    @staticmethod
    def __toot_prepare__(obj: Dict) -> Dict:
        # Pleroma has not yet converted last_status_at from datetime to date
        # so trim it here so it doesn't break when converting to date.
        # See: https://git.pleroma.social/pleroma/pleroma/-/issues/1470
        last_status_at = obj.get("last_status_at")
        if last_status_at:
            obj.update(last_status_at=obj["last_status_at"][:10])
        return obj

    @property
    def note_plaintext(self) -> str:
        return get_text(self.note)


@dataclass
class Application:
    """
    https://docs.joinmastodon.org/entities/Status/#application
    """
    name: str
    website: Optional[str]


@dataclass
class MediaAttachment:
    """
    https://docs.joinmastodon.org/entities/MediaAttachment/
    """
    id: str
    type: str
    url: str
    preview_url: str
    remote_url: Optional[str]
    meta: dict
    description: str
    blurhash: str


@dataclass
class StatusMention:
    """
    https://docs.joinmastodon.org/entities/Status/#Mention
    """
    id: str
    username: str
    url: str
    acct: str


@dataclass
class StatusTag:
    """
    https://docs.joinmastodon.org/entities/Status/#Tag
    """
    name: str
    url: str


@dataclass
class PollOption:
    """
    https://docs.joinmastodon.org/entities/Poll/#Option
    """
    title: str
    votes_count: Optional[int]


@dataclass
class Poll:
    """
    https://docs.joinmastodon.org/entities/Poll/
    """
    id: str
    expires_at: Optional[datetime]
    expired: bool
    multiple: bool
    votes_count: int
    voters_count: Optional[int]
    options: List[PollOption]
    emojis: List[CustomEmoji]
    voted: Optional[bool]
    own_votes: Optional[List[int]]


@dataclass
class PreviewCard:
    """
    https://docs.joinmastodon.org/entities/PreviewCard/
    """
    url: str
    title: str
    description: str
    type: str
    author_name: str
    author_url: str
    provider_name: str
    provider_url: str
    html: str
    width: int
    height: int
    image: Optional[str]
    embed_url: str
    blurhash: Optional[str]


@dataclass
class FilterKeyword:
    """
    https://docs.joinmastodon.org/entities/FilterKeyword/
    """
    id: str
    keyword: str
    whole_word: str


@dataclass
class FilterStatus:
    """
    https://docs.joinmastodon.org/entities/FilterStatus/
    """
    id: str
    status_id: str


@dataclass
class Filter:
    """
    https://docs.joinmastodon.org/entities/Filter/
    """
    id: str
    title: str
    context: List[str]
    expires_at: Optional[datetime]
    filter_action: str
    keywords: List[FilterKeyword]
    statuses: List[FilterStatus]


@dataclass
class FilterResult:
    """
    https://docs.joinmastodon.org/entities/FilterResult/
    """
    filter: Filter
    keyword_matches: Optional[List[str]]
    status_matches: Optional[str]


@dataclass
class Status:
    """
    https://docs.joinmastodon.org/entities/Status/
    """
    id: str
    uri: str
    created_at: datetime
    account: Account
    content: str
    visibility: str
    sensitive: bool
    spoiler_text: str
    media_attachments: List[MediaAttachment]
    application: Optional[Application]
    mentions: List[StatusMention]
    tags: List[StatusTag]
    emojis: List[CustomEmoji]
    reblogs_count: int
    favourites_count: int
    replies_count: int
    url: Optional[str]
    in_reply_to_id: Optional[str]
    in_reply_to_account_id: Optional[str]
    reblog: Optional["Status"]
    poll: Optional[Poll]
    card: Optional[PreviewCard]
    language: Optional[str]
    text: Optional[str]
    edited_at: Optional[datetime]
    favourited: Optional[bool]
    reblogged: Optional[bool]
    muted: Optional[bool]
    bookmarked: Optional[bool]
    pinned: Optional[bool]
    filtered: Optional[List[FilterResult]]

    @property
    def original(self) -> "Status":
        return self.reblog or self

    @staticmethod
    def __toot_prepare__(obj: Dict) -> Dict:
        # Pleroma has a bug where created_at is set to an empty string.
        # To avoid marking created_at as optional, which would require work
        # because we count on it always existing, set it to current datetime.
        # Possible underlying issue:
        # https://git.pleroma.social/pleroma/pleroma/-/issues/2851
        if not obj["created_at"]:
            obj["created_at"] = datetime.now().astimezone().isoformat()
        return obj


@dataclass
class Report:
    """
    https://docs.joinmastodon.org/entities/Report/
    """
    id: str
    action_taken: bool
    action_taken_at: Optional[datetime]
    category: str
    comment: str
    forwarded: bool
    created_at: datetime
    status_ids: Optional[List[str]]
    rule_ids: Optional[List[str]]
    target_account: Account


@dataclass
class Notification:
    """
    https://docs.joinmastodon.org/entities/Notification/
    """
    id: str
    type: str
    created_at: datetime
    account: Account
    status: Optional[Status]
    report: Optional[Report]


@dataclass
class InstanceUrls:
    streaming_api: str


@dataclass
class InstanceStats:
    user_count: int
    status_count: int
    domain_count: int


@dataclass
class InstanceConfigurationStatuses:
    max_characters: int
    max_media_attachments: int
    characters_reserved_per_url: int


@dataclass
class InstanceConfigurationMediaAttachments:
    supported_mime_types: List[str]
    image_size_limit: int
    image_matrix_limit: int
    video_size_limit: int
    video_frame_rate_limit: int
    video_matrix_limit: int


@dataclass
class InstanceConfigurationPolls:
    max_options: int
    max_characters_per_option: int
    min_expiration: int
    max_expiration: int


@dataclass
class InstanceConfiguration:
    """
    https://docs.joinmastodon.org/entities/V1_Instance/#configuration
    """
    statuses: InstanceConfigurationStatuses
    media_attachments: InstanceConfigurationMediaAttachments
    polls: InstanceConfigurationPolls


@dataclass
class Rule:
    """
    https://docs.joinmastodon.org/entities/Rule/
    """
    id: str
    text: str


@dataclass
class Instance:
    """
    https://docs.joinmastodon.org/entities/V1_Instance/
    """
    uri: str
    title: str
    short_description: str
    description: str
    email: str
    version: str
    urls: InstanceUrls
    stats: InstanceStats
    thumbnail: Optional[str]
    languages: List[str]
    registrations: bool
    approval_required: bool
    invites_enabled: bool
    configuration: InstanceConfiguration
    contact_account: Optional[Account]
    rules: List[Rule]


@dataclass
class Relationship:
    """
    Represents the relationship between accounts, such as following / blocking /
    muting / etc.
    https://docs.joinmastodon.org/entities/Relationship/
    """
    id: str
    following: bool
    showing_reblogs: bool
    notifying: bool
    languages: List[str]
    followed_by: bool
    blocking: bool
    blocked_by: bool
    muting: bool
    muting_notifications: bool
    requested: bool
    domain_blocking: bool
    endorsed: bool
    note: str


@dataclass
class TagHistory:
    """
    Usage statistics for given days (typically the past week).
    https://docs.joinmastodon.org/entities/Tag/#history
    """
    day: str
    uses: str
    accounts: str


@dataclass
class Tag:
    """
    Represents a hashtag used within the content of a status.
    https://docs.joinmastodon.org/entities/Tag/
    """
    name: str
    url: str
    history: List[TagHistory]
    following: Optional[bool]


@dataclass
class FeaturedTag:
    """
    Represents a hashtag that is featured on a profile.
    https://docs.joinmastodon.org/entities/FeaturedTag/
    """
    id: str
    name: str
    url: str
    statuses_count: int
    last_status_at: datetime


# Generic data class instance
T = TypeVar("T")


class ConversionError(Exception):
    """Raised when conversion fails from JSON value to data class field."""
    def __init__(
        self,
        data_class: Type,
        field_name: str,
        field_type: Type,
        field_value: Optional[str]
    ):
        super().__init__(
            f"Failed converting field `{data_class.__name__}.{field_name}` "
            + f"of type `{field_type.__name__}` from value {field_value!r}"
        )


def from_dict(cls: Type[T], data: Dict) -> T:
    """Convert a nested dict into an instance of `cls`."""
    # Apply __toot_prepare__ if it exists
    prepare = getattr(cls, '__toot_prepare__', None)
    if prepare:
        data = prepare(data)

    def _fields():
        for name, type, default in get_fields(cls):
            value = data.get(name, default)
            converted = _convert_with_error_handling(cls, name, type, value)
            yield name, converted

    return cls(**dict(_fields()))


@lru_cache(maxsize=100)
def get_fields(cls: Type) -> List[Tuple[str, Type, Any]]:
    hints = get_type_hints(cls)
    return [
        (
            field.name,
            _prune_optional(hints[field.name]),
            _get_default_value(field)
        )
        for field in dataclasses.fields(cls)
    ]


def from_dict_list(cls: Type[T], data: List[Dict]) -> List[T]:
    return [from_dict(cls, x) for x in data]


def _get_default_value(field):
    if field.default is not dataclasses.MISSING:
        return field.default

    if field.default_factory is not dataclasses.MISSING:
        return field.default_factory()

    return None


def _convert_with_error_handling(
    data_class: Type,
    field_name: str,
    field_type: Type,
    field_value: Optional[str]
):
    try:
        return _convert(field_type, field_value)
    except ConversionError:
        raise
    except Exception:
        raise ConversionError(data_class, field_name, field_type, field_value)


def _convert(field_type, value):
    if value is None:
        return None

    if field_type in [str, int, bool, dict]:
        return value

    if field_type == datetime:
        return parse_datetime(value)

    if field_type == date:
        return date.fromisoformat(value)

    if get_origin(field_type) == list:
        (inner_type,) = get_args(field_type)
        return [_convert(inner_type, x) for x in value]

    if is_dataclass(field_type):
        return from_dict(field_type, value)

    raise ValueError(f"Not implemented for type '{field_type}'")


def _prune_optional(field_type: Type) -> Type:
    """For `Optional[<type>]` returns the encapsulated `<type>`."""
    if get_origin(field_type) == Union:
        args = get_args(field_type)
        if len(args) == 2 and args[1] == type(None):  # noqa
            return args[0]

    return field_type



import logging
import subprocess
import urwid

from concurrent.futures import ThreadPoolExecutor
from typing import NamedTuple, Optional
from datetime import datetime, timezone

from toot import api, config, __version__, settings
from toot import App, User
from toot.cli import get_default_visibility
from toot.exceptions import ApiError
from toot.utils.datetime import parse_datetime

from .compose import StatusComposer
from .constants import PALETTE
from .entities import Status
from .overlays import ExceptionStackTrace, GotoMenu, Help, StatusSource, StatusLinks, StatusZoom
from .overlays import StatusDeleteConfirmation, Account
from .poll import Poll
from .timeline import Timeline
from .utils import get_max_toot_chars, parse_content_links, copy_to_clipboard
from .widgets import ModalBox

logger = logging.getLogger(__name__)

urwid.set_encoding('UTF-8')


DEFAULT_MAX_TOOT_CHARS = 500


class TuiOptions(NamedTuple):
    colors: int
    media_viewer: Optional[str]
    relative_datetimes: bool
    default_visibility: Optional[bool]


class Header(urwid.WidgetWrap):
    def __init__(self, app, user):
        self.app = app
        self.user = user

        self.text = urwid.Text("")
        self.cols = urwid.Columns([
            ("pack", urwid.Text(('header_bold', 'toot'))),
            ("pack", urwid.Text(('header', ' | {}@{}'.format(user.username, app.instance)))),
            ("pack", self.text),
        ])

        widget = urwid.AttrMap(self.cols, 'header')
        widget = urwid.Padding(widget)
        self._wrapped_widget = widget

    def clear_text(self, text):
        self.text.set_text("")

    def set_text(self, text):
        self.text.set_text(" | " + text)


class Footer(urwid.Pile):
    def __init__(self):
        self.status = urwid.Text("")
        self.message = urwid.Text("")

        return super().__init__([
            urwid.AttrMap(self.status, "footer_status"),
            urwid.AttrMap(self.message, "footer_message"),
        ])

    def set_status(self, text):
        self.status.set_text(text)

    def clear_status(self, text):
        self.status.set_text("")

    def set_message(self, text):
        self.message.set_text(text)

    def set_error_message(self, text):
        self.message.set_text(("footer_message_error", text))

    def clear_message(self):
        self.message.set_text("")


class TUI(urwid.Frame):
    """Main TUI frame."""
    loop: urwid.MainLoop
    screen: urwid.BaseScreen

    @staticmethod
    def create(app: App, user: User, args: TuiOptions):
        """Factory method, sets up TUI and an event loop."""
        screen = urwid.raw_display.Screen()
        screen.set_terminal_properties(args.colors)

        tui = TUI(app, user, screen, args)

        palette = PALETTE.copy()
        overrides = settings.get_setting("tui.palette", dict, {})
        for name, styles in overrides.items():
            palette.append(tuple([name] + styles))

        loop = urwid.MainLoop(
            tui,
            palette=palette,
            event_loop=urwid.AsyncioEventLoop(),
            unhandled_input=tui.unhandled_input,
            screen=screen,
        )
        tui.loop = loop

        return tui

    def __init__(self, app, user, screen, options: TuiOptions):
        self.app = app
        self.user = user
        self.config = config.load_config()
        self.options = options

        self.loop = None  # late init, set in `create`
        self.screen = screen
        self.executor = ThreadPoolExecutor(max_workers=1)
        self.timeline_generator = api.home_timeline_generator(app, user, limit=40)

        # Show intro screen while toots are being loaded
        self.body = self.build_intro()
        self.header = Header(app, user)
        self.footer = Footer()
        self.footer.set_status("Loading...")

        # Default max status length, updated on startup
        self.max_toot_chars = DEFAULT_MAX_TOOT_CHARS

        self.timeline = None
        self.overlay = None
        self.exception = None
        self.can_translate = False
        self.account = None
        self.followed_accounts = []
        self.preferences = {}

        super().__init__(self.body, header=self.header, footer=self.footer)

    def run(self):
        self.loop.set_alarm_in(0, lambda *args: self.async_load_instance())
        self.loop.set_alarm_in(0, lambda *args: self.async_load_preferences())
        self.loop.set_alarm_in(0, lambda *args: self.async_load_timeline(
            is_initial=True, timeline_name="home"))
        self.loop.set_alarm_in(0, lambda *args: self.async_load_followed_accounts())
        self.loop.run()
        self.executor.shutdown(wait=False)

    def build_intro(self):
        font = urwid.font.Thin6x6Font()

        # NB: Padding with width="clip" will convert the fixed BigText widget
        # to a flow widget so it can be used in a Pile.

        big_text = "Toot {}".format(__version__)
        big_text = urwid.BigText(("intro_bigtext", big_text), font)
        big_text = urwid.Padding(big_text, align="center", width="clip")

        intro = urwid.Pile([
            big_text,
            urwid.Divider(),
            urwid.Text([
                "Maintained by ",
                ("intro_smalltext", "@ihabunek"),
                " and contributors"
            ], align="center"),
            urwid.Divider(),
            urwid.Text(("intro_smalltext", "Loading toots..."), align="center"),
        ])

        return urwid.Filler(intro)

    def run_in_thread(self, fn, done_callback=None, error_callback=None):
        """Runs `fn` asynchronously in a separate thread.

        On completion calls `done_callback` if `fn` exited cleanly, or
        `error_callback` if an exception was caught. Callback methods are
        invoked in the main thread, not the thread in which `fn` is executed.
        """

        def _default_error_callback(ex):
            self.exception = ex
            self.footer.set_error_message("An exception occurred, press X to view")

        _error_callback = error_callback or _default_error_callback

        def _done(future):
            try:
                result = future.result()
                if done_callback:
                    # Use alarm to invoke callback in main thread
                    self.loop.set_alarm_in(0, lambda *args: done_callback(result))
            except Exception as ex:
                exception = ex
                logger.exception(exception)
                self.loop.set_alarm_in(0, lambda *args: _error_callback(exception))

        # TODO: replace by `self.loop.event_loop.run_in_executor` at some point
        # Added in https://github.com/urwid/urwid/issues/575
        # Not yet released at the time of this comment
        future = self.loop.event_loop._loop.run_in_executor(self.executor, fn)
        future.add_done_callback(_done)
        return future

    def connect_default_timeline_signals(self, timeline):
        urwid.connect_signal(timeline, "focus", self.refresh_footer)

    def build_timeline(self, name, statuses, local):
        def _close(*args):
            raise urwid.ExitMainLoop()

        def _next(*args):
            self.async_load_timeline(is_initial=False)

        def _toggle_save(timeline, status):
            if not timeline.name.startswith("#"):
                return
            hashtag = timeline.name[1:]
            assert isinstance(local, bool), local
            timelines = self.config.setdefault("timelines", {})
            if hashtag in timelines:
                del timelines[hashtag]
                self.footer.set_message("#{} unpinned".format(hashtag))
            else:
                timelines[hashtag] = {"local": local}
                self.footer.set_message("#{} pinned".format(hashtag))
            self.loop.set_alarm_in(5, lambda *args: self.footer.clear_message())
            config.save_config(self.config)

        timeline = Timeline(self, name, statuses)

        self.connect_default_timeline_signals(timeline)
        urwid.connect_signal(timeline, "next", _next)
        urwid.connect_signal(timeline, "close", _close)
        urwid.connect_signal(timeline, "save", _toggle_save)

        return timeline

    def make_status(self, status_data):
        is_mine = self.user.username == status_data["account"]["acct"]
        return Status(status_data, is_mine, self.app.instance)

    def show_thread(self, status):
        def _close(*args):
            """When thread is closed, go back to the main timeline."""
            self.body = self.timeline
            self.body.refresh_status_details()
            self.refresh_footer(self.timeline)

        # This is pretty fast, so it's probably ok to block while context is
        # loaded, can be made async later if needed
        context = api.context(self.app, self.user, status.original.id).json()
        ancestors = [self.make_status(s) for s in context["ancestors"]]
        descendants = [self.make_status(s) for s in context["descendants"]]
        statuses = ancestors + [status] + descendants
        focus = len(ancestors)

        timeline = Timeline(self, "thread", statuses, focus=focus, is_thread=True)

        self.connect_default_timeline_signals(timeline)
        urwid.connect_signal(timeline, "close", _close)
        self.body = timeline
        timeline.refresh_status_details()
        self.refresh_footer(timeline)

    def async_load_timeline(self, is_initial, timeline_name=None, local=None):
        """Asynchronously load a list of statuses."""

        def _load_statuses():
            self.footer.set_message("Loading statuses...")
            try:
                data = next(self.timeline_generator)
            except StopIteration:
                return []
            finally:
                self.footer.clear_message()

            return [self.make_status(s) for s in data]

        def _done_initial(statuses):
            """Process initial batch of statuses, construct a Timeline."""
            self.timeline = self.build_timeline(timeline_name, statuses, local)
            self.timeline.refresh_status_details()  # Draw first status
            self.refresh_footer(self.timeline)
            self.body = self.timeline

        def _done_next(statuses):
            """Process sequential batch of statuses, adds statuses to the
            existing timeline."""
            self.timeline.append_statuses(statuses)

        return self.run_in_thread(_load_statuses,
            done_callback=_done_initial if is_initial else _done_next)

    def async_load_instance(self):
        """
        Attempt to update max_toot_chars from instance data.
        Does not work on vanilla Mastodon, works on Pleroma.
        See: https://github.com/tootsuite/mastodon/issues/4915

        Also attempt to update translation flag from instance
        data. Translation is only present on Mastodon 4+ servers
        where the administrator has enabled this feature.
        See: https://github.com/mastodon/mastodon/issues/19328
        """
        def _load_instance():
            return api.get_instance(self.app.base_url).json()

        def _done(instance):
            self.max_toot_chars = get_max_toot_chars(instance, DEFAULT_MAX_TOOT_CHARS)
            logger.info(f"Max toot chars set to: {self.max_toot_chars}")

            if "translation" in instance:
                # instance is advertising translation service
                self.can_translate = instance["translation"]["enabled"]
            elif "version" in instance:
                # fallback check:
                # get the major version number of the server
                # this works for Mastodon and Pleroma version strings
                # Mastodon versions < 4 do not have translation service
                # Revisit this logic if Pleroma implements translation
                ch = instance["version"][0]
                self.can_translate = int(ch) > 3 if ch.isnumeric() else False

        return self.run_in_thread(_load_instance, done_callback=_done)

    def async_load_preferences(self):
        """
        Attempt to update user preferences from instance.
        https://docs.joinmastodon.org/methods/preferences/
        """
        def _load_preferences():
            return api.get_preferences(self.app, self.user).json()

        def _done(preferences):
            self.preferences = preferences

        return self.run_in_thread(_load_preferences, done_callback=_done)

    def async_load_followed_accounts(self):
        def _load_accounts():
            try:
                acct = f'@{self.user.username}@{self.user.instance}'
                self.account = api.find_account(self.app, self.user, acct)
                return api.following(self.app, self.user, self.account["id"])
            except ApiError:
                # not supported by all Mastodon servers so fail silently if necessary
                return []

        def _done_accounts(accounts):
            self.followed_accounts = {a["acct"] for a in accounts}

        self.run_in_thread(_load_accounts, done_callback=_done_accounts)

    def refresh_footer(self, timeline):
        """Show status details in footer."""
        status, index, count = timeline.get_focused_status_with_counts()
        self.footer.set_status([
            ("footer_status_bold", "[{}] ".format(timeline.name)),
        ] + ([status.id, " - status ", str(index + 1), " of ", str(count)]
            if status else ["no focused status"]))

    def show_status_source(self, status):
        self.open_overlay(
            widget=StatusSource(status),
            title="Status source",
        )

    def clear_screen(self):
        self.screen.clear()

    def show_links(self, status):
        links = parse_content_links(status.original.data["content"]) if status else []
        post_attachments = status.original.data["media_attachments"] or []
        reblog_attachments = (status.data["reblog"]["media_attachments"] if status.data["reblog"] else None) or []

        for a in post_attachments + reblog_attachments:
            url = a["remote_url"] or a["url"]
            links.append((url, a["description"] if a["description"] else url))

        def _clear(*args):
            self.clear_screen()

        if links:
            links = list(set(links))  # deduplicate links
            links = sorted(links, key=lambda link: link[0])  # sort alphabetically by URL
            sl_widget = StatusLinks(links)
            urwid.connect_signal(sl_widget, "clear-screen", _clear)
            self.open_overlay(
                widget=sl_widget,
                title="Status links",
                options={"height": len(links) + 2},
            )

    def show_status_zoom(self, status_details):
        self.open_overlay(
            widget=StatusZoom(status_details),
            title="Status zoom",
        )

    def show_exception(self, exception):
        self.open_overlay(
            widget=ExceptionStackTrace(exception),
            title="Unhandled Exception",
        )

    def show_compose(self, in_reply_to=None):
        def _close(*args):
            self.close_overlay()

        def _post(timeline, *args):
            self.post_status(*args)

        # If the user specified --default-visibility, use that; otherwise,
        # try to use the server-side default visibility.  If that fails, fall
        # back to get_default_visibility().
        visibility = (self.options.default_visibility or
                      self.preferences.get('posting:default:visibility',
                                           get_default_visibility()))

        composer = StatusComposer(self.max_toot_chars, self.user.username,
                                  visibility, in_reply_to)
        urwid.connect_signal(composer, "close", _close)
        urwid.connect_signal(composer, "post", _post)
        self.open_overlay(composer, title="Compose status")

    def async_edit(self, status):
        def _fetch_source():
            return api.fetch_status_source(self.app, self.user, status.id).json()

        def _done(source):
            self.close_overlay()
            self.show_edit(status, source)

        please_wait = ModalBox("Loading status...")
        self.open_overlay(please_wait)

        self.run_in_thread(_fetch_source, done_callback=_done)

    def show_edit(self, status, source):
        def _close(*args):
            self.close_overlay()

        def _edit(timeline, *args):
            self.edit_status(status, *args)

        composer = StatusComposer(self.max_toot_chars, self.user.username,
                                  visibility=None, edit=status, source=source)
        urwid.connect_signal(composer, "close", _close)
        urwid.connect_signal(composer, "post", _edit)
        self.open_overlay(composer, title="Edit status")

    def show_goto_menu(self):
        user_timelines = self.config.get("timelines", {})
        user_lists = api.get_lists(self.app, self.user) or []

        menu = GotoMenu(user_timelines, user_lists)
        urwid.connect_signal(menu, "home_timeline",
            lambda x: self.goto_home_timeline())
        urwid.connect_signal(menu, "public_timeline",
            lambda x, local: self.goto_public_timeline(local))
        urwid.connect_signal(menu, "bookmark_timeline",
            lambda x, local: self.goto_bookmarks())
        urwid.connect_signal(menu, "notification_timeline",
            lambda x, local: self.goto_notifications())
        urwid.connect_signal(menu, "conversation_timeline",
            lambda x, local: self.goto_conversations())
        urwid.connect_signal(menu, "personal_timeline",
            lambda x, local: self.goto_personal_timeline())
        urwid.connect_signal(menu, "hashtag_timeline",
            lambda x, tag, local: self.goto_tag_timeline(tag, local=local))
        urwid.connect_signal(menu, "list_timeline",
            lambda x, list_item: self.goto_list_timeline(list_item))

        self.open_overlay(menu, title="Go to", options=dict(
            align="center", width=("relative", 60),
            valign="middle", height=18 + len(user_timelines) + len(user_lists),
        ))

    def show_help(self):
        self.open_overlay(Help(), title="Help")

    def show_poll(self, status):
        self.open_overlay(
            widget=Poll(self.app, self.user, status),
            title="Poll",
        )

    def goto_home_timeline(self):
        self.timeline_generator = api.home_timeline_generator(
            self.app, self.user, limit=40)
        promise = self.async_load_timeline(is_initial=True, timeline_name="home")
        promise.add_done_callback(lambda *args: self.close_overlay())

    def goto_public_timeline(self, local):
        self.timeline_generator = api.public_timeline_generator(
            self.app, self.user, local=local, limit=40)
        timeline_name = "local public" if local else "global public"
        promise = self.async_load_timeline(is_initial=True, timeline_name=timeline_name)
        promise.add_done_callback(lambda *args: self.close_overlay())

    def goto_bookmarks(self):
        self.timeline_generator = api.bookmark_timeline_generator(
            self.app, self.user, limit=40)
        promise = self.async_load_timeline(is_initial=True, timeline_name="bookmarks")
        promise.add_done_callback(lambda *args: self.close_overlay())

    def goto_notifications(self):
        self.timeline_generator = api.notification_timeline_generator(
            self.app, self.user, limit=40)
        promise = self.async_load_timeline(is_initial=True, timeline_name="notifications")
        promise.add_done_callback(lambda *args: self.close_overlay())

    def goto_conversations(self):
        self.timeline_generator = api.conversation_timeline_generator(
            self.app, self.user, limit=40
        )
        promise = self.async_load_timeline(
            is_initial=True, timeline_name="conversations"
        )
        promise.add_done_callback(lambda *args: self.close_overlay())

    def goto_tag_timeline(self, tag, local):
        self.timeline_generator = api.tag_timeline_generator(
            self.app, self.user, tag, local=local, limit=40)
        promise = self.async_load_timeline(
            is_initial=True, timeline_name="#{}".format(tag), local=local,
        )
        promise.add_done_callback(lambda *args: self.close_overlay())

    def goto_personal_timeline(self):
        account_name = f"{self.user.username}@{self.user.instance}"

        self.timeline_generator = api.account_timeline_generator(
            self.app, self.user, account_name, reblogs=True, limit=40)
        promise = self.async_load_timeline(is_initial=True, timeline_name=f"personal {account_name}")
        promise.add_done_callback(lambda *args: self.close_overlay())

    def goto_list_timeline(self, list_item):
        self.timeline_generator = api.timeline_list_generator(
            self.app, self.user, list_item['id'], limit=40)
        promise = self.async_load_timeline(
            is_initial=True, timeline_name=f"\N{clipboard}{list_item['title']}")
        promise.add_done_callback(lambda *args: self.close_overlay())

    def show_media(self, status):
        urls = [m["url"] for m in status.original.data["media_attachments"]]
        if not urls:
            return

        media_viewer = self.options.media_viewer
        if media_viewer:
            try:
                subprocess.run([media_viewer] + urls)
            except FileNotFoundError:
                self.footer.set_error_message(f"Media viewer not found: '{media_viewer}'")
            except Exception as ex:
                self.exception = ex
                self.footer.set_error_message("Failed invoking media viewer. Press X to see exception.")
        else:
            self.footer.set_error_message("Media viewer not configured")

    def show_context_menu(self, status):
        # TODO: show context menu
        pass

    def show_delete_confirmation(self, status):
        def _delete(widget):
            promise = self.async_delete_status(self.timeline, status)
            promise.add_done_callback(lambda *args: self.close_overlay())

        def _close(widget):
            self.close_overlay()

        widget = StatusDeleteConfirmation(status)
        urwid.connect_signal(widget, "close", _close)
        urwid.connect_signal(widget, "delete", _delete)
        self.open_overlay(widget, title="Delete status?", options=dict(
            align="center", width=30,
            valign="middle", height=4,
        ))

    def post_status(self, content, warning, visibility, in_reply_to_id):
        data = api.post_status(
            self.app,
            self.user,
            content,
            spoiler_text=warning,
            visibility=visibility,
            in_reply_to_id=in_reply_to_id
        ).json()

        status = self.make_status(data)

        # TODO: fetch new items from the timeline?

        self.footer.set_message("Status posted {} \\o/".format(status.id))
        self.close_overlay()

    def edit_status(self, status, content, warning, visibility, in_reply_to_id):
        # We don't support editing polls (yet), so to avoid losing the poll
        # data from the original toot, copy it to the edit request.
        poll_args = {}
        poll = status.original.data.get('poll', None)

        if poll is not None:
            poll_args['poll_options'] = [o['title'] for o in poll['options']]
            poll_args['poll_multiple'] = poll['multiple']

            # Convert absolute expiry time into seconds from now.
            expires_at = parse_datetime(poll['expires_at'])
            expires_in = int((expires_at - datetime.now(timezone.utc)).total_seconds())
            poll_args['poll_expires_in'] = expires_in

            if 'hide_totals' in poll:
                poll_args['poll_hide_totals'] = poll['hide_totals']

        data = api.edit_status(
            self.app,
            self.user,
            status.id,
            content,
            spoiler_text=warning,
            visibility=visibility,
            **poll_args
        ).json()

        new_status = self.make_status(data)

        self.footer.set_message("Status edited {} \\o/".format(status.id))
        self.close_overlay()

        if self.timeline is not None:
            self.timeline.update_status(new_status)

    def show_account(self, account_id):
        account = api.whois(self.app, self.user, account_id)
        relationship = api.get_relationship(self.app, self.user, account_id)
        self.open_overlay(
            widget=Account(self.app, self.user, account, relationship),
            title="Account",
        )

    def async_toggle_favourite(self, timeline, status):
        def _favourite():
            api.favourite(self.app, self.user, status.id)

        def _unfavourite():
            api.unfavourite(self.app, self.user, status.id)

        def _done(loop):
            # Create a new Status with flipped favourited flag
            new_data = status.data
            new_data["favourited"] = not status.favourited
            new_status = self.make_status(new_data)
            timeline.update_status(new_status)

        self.run_in_thread(
            _unfavourite if status.favourited else _favourite,
            done_callback=_done
        )

    def async_toggle_reblog(self, timeline, status):
        def _reblog():
            api.reblog(self.app, self.user, status.original.id, visibility=get_default_visibility())

        def _unreblog():
            api.unreblog(self.app, self.user, status.original.id)

        def _done(loop):
            # Create a new Status with flipped reblogged flag
            new_data = status.data
            new_status = self.make_status(new_data)
            new_status.original.reblogged = not status.original.reblogged
            timeline.update_status(new_status)

        # Check if status is rebloggable
        no_reblog_because_private = status.visibility == "private" and not status.is_mine
        no_reblog_because_direct = status.visibility == "direct"
        if no_reblog_because_private or no_reblog_because_direct:
            self.footer.set_error_message("You may not reblog this {} status".format(status.visibility))
            return

        self.run_in_thread(
            _unreblog if status.original.reblogged else _reblog,
            done_callback=_done
        )

    def async_translate(self, timeline, status):
        def _translate():
            self.footer.set_message("Translating status {}".format(status.original.id))

            try:
                response = api.translate(self.app, self.user, status.original.id)
                if response["content"]:
                    self.footer.set_message("Status translated")
                else:
                    self.footer.set_error_message("Server returned empty translation")
                    response = None
            except Exception:
                response = None
                self.footer.set_error_message("Translate server error")

            self.loop.set_alarm_in(3, lambda *args: self.footer.clear_message())
            return response

        def _done(response):
            if response is not None:
                status.original.translation = response["content"]
                status.original.translated_from = response["detected_source_language"]
                status.original.show_translation = True
                timeline.update_status(status)

        # If already translated, toggle showing translation
        if status.original.translation:
            status.original.show_translation = not status.original.show_translation
            timeline.update_status(status)
        else:
            self.run_in_thread(_translate, done_callback=_done)

    def async_toggle_bookmark(self, timeline, status):
        def _bookmark():
            api.bookmark(self.app, self.user, status.id)

        def _unbookmark():
            api.unbookmark(self.app, self.user, status.id)

        def _done(loop):
            # Create a new Status with flipped bookmarked flag
            new_data = status.data
            new_data["bookmarked"] = not status.bookmarked
            new_status = self.make_status(new_data)
            timeline.update_status(new_status)

        self.run_in_thread(
            _unbookmark if status.bookmarked else _bookmark,
            done_callback=_done
        )

    def async_delete_status(self, timeline, status):
        def _delete():
            api.delete_status(self.app, self.user, status.id)

        def _done(loop):
            timeline.remove_status(status)

        return self.run_in_thread(_delete, done_callback=_done)

    def copy_status(self, status):
        # TODO: copy a better version of status content
        # including URLs
        copy_to_clipboard(self.screen, status.original.data["content"])
        self.footer.set_message(f"Status {status.original.id} copied")

    # --- Overlay handling -----------------------------------------------------

    default_overlay_options = dict(
        align="center", width=("relative", 80),
        valign="middle", height=("relative", 80),
    )

    def open_overlay(self, widget, options={}, title=""):
        top_widget = urwid.LineBox(widget, title=title)
        bottom_widget = self.body

        _options = self.default_overlay_options.copy()
        _options.update(options)

        self.overlay = urwid.Overlay(
            top_widget,
            bottom_widget,
            **_options
        )
        self.body = self.overlay

    def close_overlay(self):
        self.body = self.overlay.bottom_w
        self.overlay = None
        if self.timeline:
            self.timeline.refresh_status_details()

    def refresh_timeline(self):
        # No point in refreshing the bookmarks timeline
        # and we don't have a good way to refresh a
        # list timeline yet (no reference to list ID kept)
        if (not self.timeline
                or self.timeline.name == 'bookmarks'
                or self.timeline.name.startswith("\N{clipboard}")):
            return

        if self.timeline.name.startswith("#"):
            self.timeline_generator = api.tag_timeline_generator(
                self.app, self.user, self.timeline.name[1:], limit=40)
        elif self.timeline.name.startswith("\N{clipboard}"):
            self.timeline_generator = api.tag_timeline_generator(
                self.app, self.user, self.timeline.name[1:], limit=40)
        else:
            if self.timeline.name.endswith("public"):
                self.timeline_generator = api.public_timeline_generator(
                    self.app, self.user, local=self.timeline.name.startswith("local"), limit=40)
            elif self.timeline.name == "notifications":
                self.timeline_generator = api.notification_timeline_generator(
                    self.app, self.user, limit=40)
            elif self.timeline.name == "conversations":
                self.timeline_generator = api.conversation_timeline_generator(
                    self.app, self.user, limit=40)
            else:
                # default to home timeline
                self.timeline_generator = api.home_timeline_generator(
                    self.app, self.user, limit=40)

        self.async_load_timeline(is_initial=True, timeline_name=self.timeline.name)

    # --- Keys -----------------------------------------------------------------

    def unhandled_input(self, key):
        # TODO: this should not be in unhandled input
        if key in ('x', 'X'):
            if self.exception:
                self.show_exception(self.exception)

        elif key in ('g', 'G'):
            if not self.overlay:
                self.show_goto_menu()

        elif key == '?':
            if not self.overlay:
                self.show_help()

        elif key == ',':
            if not self.overlay:
                self.refresh_timeline()

        elif key == 'esc':
            if self.overlay:
                self.close_overlay()
            elif self.timeline.name != "home":
                # similar to goto_home_timeline() but without handling overlay (absent here)
                self.timeline_generator = api.home_timeline_generator(
                    self.app, self.user, limit=40)
                self.async_load_timeline(is_initial=True, timeline_name="home")

        elif key in ('q', 'Q'):
            if self.overlay:
                self.close_overlay()
            else:
                raise urwid.ExitMainLoop()



import click
import json as pyjson

from itertools import chain
from typing import Optional

from toot import api
from toot.cli.validators import validate_instance
from toot.entities import Instance, Status, from_dict, Account
from toot.exceptions import ApiError, ConsoleError
from toot.output import print_account, print_instance, print_search_results, print_status, print_timeline
from toot.cli import InstanceParamType, cli, get_context, json_option, pass_context, Context


@cli.command()
@json_option
@pass_context
def whoami(ctx: Context, json: bool):
    """Display logged in user details"""
    response = api.verify_credentials(ctx.app, ctx.user)

    if json:
        click.echo(response.text)
    else:
        account = from_dict(Account, response.json())
        print_account(account)


@cli.command()
@click.argument("account")
@json_option
@pass_context
def whois(ctx: Context, account: str, json: bool):
    """Display account details"""
    account_dict = api.find_account(ctx.app, ctx.user, account)

    # Here it's not possible to avoid parsing json since it's needed to find the account.
    if json:
        click.echo(pyjson.dumps(account_dict))
    else:
        account_obj = from_dict(Account, account_dict)
        print_account(account_obj)


@cli.command()
@click.argument("instance", type=InstanceParamType(), callback=validate_instance, required=False)
@json_option
def instance(instance: Optional[str], json: bool):
    """Display instance details

    INSTANCE can be a domain or base URL of the instance to display.
    e.g. 'mastodon.social' or 'https://mastodon.social'. If not
    given will display details for the currently logged in instance.
    """
    if not instance:
        context = get_context()
        if not context.app:
            raise click.ClickException("INSTANCE argument not given and not logged in")
        instance = context.app.base_url

    try:
        response = api.get_instance(instance)
    except ApiError:
        raise ConsoleError(
            f"Instance not found at {instance}.\n" +
            "The given domain probably does not host a Mastodon instance."
        )

    if json:
        click.echo(response.text)
    else:
        print_instance(from_dict(Instance, response.json()))


@cli.command()
@click.argument("query")
@click.option("-r", "--resolve", is_flag=True, help="Resolve non-local accounts")
@json_option
@pass_context
def search(ctx: Context, query: str, resolve: bool, json: bool):
    """Search for users or hashtags"""
    response = api.search(ctx.app, ctx.user, query, resolve)
    if json:
        click.echo(response.text)
    else:
        print_search_results(response.json())


@cli.command()
@click.argument("status_id")
@json_option
@pass_context
def status(ctx: Context, status_id: str, json: bool):
    """Show a single status"""
    response = api.fetch_status(ctx.app, ctx.user, status_id)
    if json:
        click.echo(response.text)
    else:
        status = from_dict(Status, response.json())
        print_status(status)


@cli.command()
@click.argument("status_id")
@json_option
@pass_context
def thread(ctx: Context, status_id: str, json: bool):
    """Show thread for a toot."""
    context_response = api.context(ctx.app, ctx.user, status_id)
    if json:
        click.echo(context_response.text)
    else:
        toot = api.fetch_status(ctx.app, ctx.user, status_id).json()
        context = context_response.json()

        statuses = chain(context["ancestors"], [toot], context["descendants"])
        print_timeline(from_dict(Status, s) for s in statuses)



# Color definitions are tuples of:
#   - name
#   - foreground (normal mode)
#   - background (normal mode)
#   - foreground (monochrome mode)
#   - foreground (high color mode)
#   - background (high color mode)
#
# See:
# http://urwid.org/tutorial/index.html#display-attributes
# http://urwid.org/manual/displayattributes.html#using-display-attributes

PALETTE = [
    # Components
    ('button', 'white', 'black'),
    ('button_focused', 'light gray', 'dark magenta', 'bold,underline'),
    ('card_author', 'yellow', ''),
    ('card_title', 'dark green', ''),
    ('columns_divider', 'white', 'dark blue'),
    ('content_warning', 'white', 'dark magenta'),
    ('editbox', 'white', 'black'),
    ('editbox_focused', 'white', 'dark magenta'),
    ('footer_message', 'dark green', ''),
    ('footer_message_error', 'light red', ''),
    ('footer_status', 'white', 'dark blue'),
    ('footer_status_bold', 'white, bold', 'dark blue'),
    ('header', 'white', 'dark blue'),
    ('header_bold', 'white,bold', 'dark blue', 'bold'),
    ('intro_bigtext', 'yellow', ''),
    ('intro_smalltext', 'light blue', ''),
    ('poll_bar', 'white', 'dark blue'),
    ('status_detail_account', 'dark green', ''),
    ('status_detail_bookmarked', 'light red', ''),
    ('status_detail_timestamp', 'light blue', ''),
    ('status_list_account', 'dark green', ''),
    ('status_list_selected', 'white,bold', 'dark green', 'bold,underline'),
    ('status_list_timestamp', 'light blue', ''),

    # Functional
    ('account', 'dark green', ''),
    ('hashtag', 'light cyan,bold', '', 'bold'),
    ('hashtag_followed', 'yellow,bold', '', 'bold'),
    ('link', ',italics', '', ',italics'),
    ('link_focused', ',italics', 'dark magenta', "underline,italics"),
    ('shortcut', 'light blue', ''),
    ('shortcut_highlight', 'white,bold', '', 'bold'),
    ('warning', 'light red', ''),

    # Visiblity
    ('visibility_public', 'dark gray', ''),
    ('visibility_unlisted', 'white', ''),
    ('visibility_private', 'dark cyan', ''),
    ('visibility_direct', 'yellow', ''),

    # Styles
    ('bold', ',bold', ''),
    ('dim', 'dark gray', ''),
    ('highlight', 'yellow', ''),
    ('success', 'dark green', ''),

    # HTML tag styling
    ('a', ',italics', '', 'italics'),
    # em tag is mapped to i
    ('i', ',italics', '', 'italics'),
    # strong tag is mapped to b
    ('b', ',bold', '', 'bold'),
    # special case for bold + italic nested tags
    ('bi', ',bold,italics', '', ',bold,italics'),
    ('u', ',underline', '', ',underline'),
    ('del', ',strikethrough', '', ',strikethrough'),
    ('code', 'light gray, standout', '', ',standout'),
    ('pre', 'light gray, standout', '', ',standout'),
    ('blockquote', 'light gray', '', ''),
    ('h1', ',bold', '', ',bold'),
    ('h2', ',bold', '', ',bold'),
    ('h3', ',bold', '', ',bold'),
    ('h4', ',bold', '', ',bold'),
    ('h5', ',bold', '', ',bold'),
    ('h6', ',bold', '', ',bold'),
    ('class_mention_hashtag', 'light cyan', '', ''),
    ('class_hashtag', 'light cyan', '', ''),

]

VISIBILITY_OPTIONS = [
    ("public", "Public", "Post to public timelines"),
    ("unlisted", "Unlisted", "Do not post to public timelines"),
    ("private", "Private", "Post to followers only"),
    ("direct", "Direct", "Post to mentioned users only"),
]



import re
from typing import List

from toot import api, cli
from toot.entities import FeaturedTag, Tag, from_dict, from_dict_list


def test_tags(run):
    result = run(cli.tags.tags, "followed")
    assert result.exit_code == 0
    assert result.stdout.strip() == "You're not following any hashtags"

    result = run(cli.tags.tags, "follow", "foo")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ You are now following #foo"

    result = run(cli.tags.tags, "followed")
    assert result.exit_code == 0
    assert _find_tags(result.stdout) == ["#foo"]

    result = run(cli.tags.tags, "follow", "bar")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ You are now following #bar"

    result = run(cli.tags.tags, "followed")
    assert result.exit_code == 0
    assert _find_tags(result.stdout) == ["#bar", "#foo"]

    result = run(cli.tags.tags, "unfollow", "foo")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ You are no longer following #foo"

    result = run(cli.tags.tags, "followed")
    assert result.exit_code == 0
    assert _find_tags(result.stdout) == ["#bar"]

    result = run(cli.tags.tags, "unfollow", "bar")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ You are no longer following #bar"

    result = run(cli.tags.tags, "followed")
    assert result.exit_code == 0
    assert result.stdout.strip() == "You're not following any hashtags"


def test_tags_json(run_json):
    result = run_json(cli.tags.tags, "followed", "--json")
    assert result == []

    result = run_json(cli.tags.tags, "follow", "foo", "--json")
    tag = from_dict(Tag, result)
    assert tag.name == "foo"
    assert tag.following is True

    result = run_json(cli.tags.tags, "followed", "--json")
    [tag] = from_dict_list(Tag, result)
    assert tag.name == "foo"
    assert tag.following is True

    result = run_json(cli.tags.tags, "follow", "bar", "--json")
    tag = from_dict(Tag, result)
    assert tag.name == "bar"
    assert tag.following is True

    result = run_json(cli.tags.tags, "followed", "--json")
    tags = from_dict_list(Tag, result)
    [bar, foo] = sorted(tags, key=lambda t: t.name)
    assert foo.name == "foo"
    assert foo.following is True
    assert bar.name == "bar"
    assert bar.following is True

    result = run_json(cli.tags.tags, "unfollow", "foo", "--json")
    tag = from_dict(Tag, result)
    assert tag.name == "foo"
    assert tag.following is False

    result = run_json(cli.tags.tags, "unfollow", "bar", "--json")
    tag = from_dict(Tag, result)
    assert tag.name == "bar"
    assert tag.following is False

    result = run_json(cli.tags.tags, "followed", "--json")
    assert result == []


def test_tags_featured(run, app, user):
    result = run(cli.tags.tags, "featured")
    assert result.exit_code == 0
    assert result.stdout.strip() == "You don't have any featured hashtags"

    result = run(cli.tags.tags, "feature", "foo")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Tag #foo is now featured"

    result = run(cli.tags.tags, "featured")
    assert result.exit_code == 0
    assert _find_tags(result.stdout) == ["#foo"]

    result = run(cli.tags.tags, "feature", "bar")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Tag #bar is now featured"

    result = run(cli.tags.tags, "featured")
    assert result.exit_code == 0
    assert _find_tags(result.stdout) == ["#bar", "#foo"]

    # Unfeature by Name
    result = run(cli.tags.tags, "unfeature", "foo")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Tag #foo is no longer featured"

    result = run(cli.tags.tags, "featured")
    assert result.exit_code == 0
    assert _find_tags(result.stdout) == ["#bar"]

    # Unfeature by ID
    tag = api.find_featured_tag(app, user, "bar")
    assert tag is not None

    result = run(cli.tags.tags, "unfeature", tag["id"])
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Tag #bar is no longer featured"

    result = run(cli.tags.tags, "featured")
    assert result.exit_code == 0
    assert result.stdout.strip() == "You don't have any featured hashtags"


def test_tags_featured_json(run_json):
    result = run_json(cli.tags.tags, "featured", "--json")
    assert result == []

    result = run_json(cli.tags.tags, "feature", "foo", "--json")
    tag = from_dict(FeaturedTag, result)
    assert tag.name == "foo"

    result = run_json(cli.tags.tags, "featured", "--json")
    [tag] = from_dict_list(FeaturedTag, result)
    assert tag.name == "foo"

    result = run_json(cli.tags.tags, "feature", "bar", "--json")
    tag = from_dict(FeaturedTag, result)
    assert tag.name == "bar"

    result = run_json(cli.tags.tags, "featured", "--json")
    tags = from_dict_list(FeaturedTag, result)
    [bar, foo] = sorted(tags, key=lambda t: t.name)
    assert foo.name == "foo"
    assert bar.name == "bar"

    result = run_json(cli.tags.tags, "unfeature", "foo", "--json")
    assert result == {}

    result = run_json(cli.tags.tags, "unfeature", "bar", "--json")
    assert result == {}

    result = run_json(cli.tags.tags, "featured", "--json")
    assert result == []


def _find_tags(txt: str) -> List[str]:
    return sorted(re.findall(r"#\w+", txt))



from urwid import Divider, Filler, Pile
from toot.tui.richtext import url_to_widget
from urwidgets import Hyperlink, TextEmbed

from toot.tui.richtext.richtext import html_to_widgets


def test_url_to_widget():
    url = "http://foo.bar"
    embed_widget = url_to_widget(url)
    assert isinstance(embed_widget, TextEmbed)

    [(filler, length)] = embed_widget.embedded
    assert length == len(url)
    assert isinstance(filler, Filler)

    link_widget = filler.base_widget
    assert isinstance(link_widget, Hyperlink)

    assert link_widget.attrib == "link"
    assert link_widget.text == url
    assert link_widget.uri == url


def test_html_to_widgets():
    html = """
    <p>foo</p>
    <p>foo <b>bar</b> <i>baz</i></p>
    """.strip()

    [foo, divider, bar] = html_to_widgets(html)

    assert isinstance(foo, Pile)
    assert isinstance(divider, Divider)
    assert isinstance(bar, Pile)

    [(foo_embed, _)] = foo.contents
    assert foo_embed.embedded == []
    assert foo_embed.attrib == []
    assert foo_embed.text == "foo"

    [(bar_embed, _)] = bar.contents
    assert bar_embed.embedded == []
    assert bar_embed.attrib == [(None, 4), ("b", 3), (None, 1), ("i", 3)]
    assert bar_embed.text == "foo bar baz"



import math
import os

from datetime import datetime, timezone


def parse_datetime(value):
    """Returns an aware datetime in local timezone"""
    dttm = datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f%z")

    # When running tests return datetime in UTC so that tests don't depend on
    # the local timezone
    if "PYTEST_CURRENT_TEST" in os.environ:
        return dttm.astimezone(timezone.utc)

    return dttm.astimezone()


SECOND = 1
MINUTE = SECOND * 60
HOUR = MINUTE * 60
DAY = HOUR * 24
WEEK = DAY * 7


def time_ago(value: datetime) -> str:
    now = datetime.now().astimezone()
    delta = now.timestamp() - value.timestamp()

    if delta < 1:
        return "now"

    if delta < 8 * DAY:
        if delta < MINUTE:
            return f"{math.floor(delta / SECOND)}".rjust(2, " ") + "s"
        if delta < HOUR:
            return f"{math.floor(delta / MINUTE)}".rjust(2, " ") + "m"
        if delta < DAY:
            return f"{math.floor(delta / HOUR)}".rjust(2, " ") + "h"
        return f"{math.floor(delta / DAY)}".rjust(2, " ") + "d"

    if delta < 53 * WEEK:  # not exactly correct but good enough as a boundary
        return f"{math.floor(delta / WEEK)}".rjust(2, " ") + "w"

    return ">1y"



import pytest

from uuid import uuid4
from tests.utils import run_with_retries

from toot import api, cli
from toot.entities import from_dict, Status
from tests.integration.conftest import TOOT_TEST_BASE_URL, register_account


# TODO: If fixture is not overriden here, tests fail, not sure why, figure it out
@pytest.fixture(scope="module")
def user(app):
    return register_account(app)


@pytest.fixture(scope="module")
def other_user(app):
    return register_account(app)


@pytest.fixture(scope="module")
def friend_user(app, user):
    friend = register_account(app)
    friend_account = api.find_account(app, user, friend.username)
    api.follow(app, user, friend_account["id"])
    return friend


@pytest.fixture(scope="module")
def friend_list(app, user, friend_user):
    friend_account = api.find_account(app, user, friend_user.username)
    list = api.create_list(app, user, str(uuid4())).json()
    api.add_accounts_to_list(app, user, list["id"], account_ids=[friend_account["id"]])
    return list


def test_timelines(app, user, other_user, friend_user, friend_list, run):
    status1 = _post_status(app, user, "#foo")
    status2 = _post_status(app, other_user, "#bar")
    status3 = _post_status(app, friend_user, "#foo #bar")

    # Home timeline
    def test_home():
        result = run(cli.timelines.timeline)
        assert result.exit_code == 0
        assert status1.id in result.stdout
        assert status2.id not in result.stdout
        assert status3.id in result.stdout
    run_with_retries(test_home)

    # Public timeline
    result = run(cli.timelines.timeline, "--public")
    assert result.exit_code == 0
    assert status1.id in result.stdout
    assert status2.id in result.stdout
    assert status3.id in result.stdout

    # Anon public timeline
    result = run(cli.timelines.timeline, "--instance", TOOT_TEST_BASE_URL, "--public")
    assert result.exit_code == 0
    assert status1.id in result.stdout
    assert status2.id in result.stdout
    assert status3.id in result.stdout

    # Tag timeline
    result = run(cli.timelines.timeline, "--tag", "foo")
    assert result.exit_code == 0
    assert status1.id in result.stdout
    assert status2.id not in result.stdout
    assert status3.id in result.stdout

    result = run(cli.timelines.timeline, "--tag", "bar")
    assert result.exit_code == 0
    assert status1.id not in result.stdout
    assert status2.id in result.stdout
    assert status3.id in result.stdout

    # Anon tag timeline
    result = run(cli.timelines.timeline, "--instance", TOOT_TEST_BASE_URL, "--tag", "foo")
    assert result.exit_code == 0
    assert status1.id in result.stdout
    assert status2.id not in result.stdout
    assert status3.id in result.stdout

    # List timeline (by list name)
    result = run(cli.timelines.timeline, "--list", friend_list["title"])
    assert result.exit_code == 0
    assert status1.id not in result.stdout
    assert status2.id not in result.stdout
    assert status3.id in result.stdout

    # List timeline (by list ID)
    result = run(cli.timelines.timeline, "--list", friend_list["id"])
    assert result.exit_code == 0
    assert status1.id not in result.stdout
    assert status2.id not in result.stdout
    assert status3.id in result.stdout

    # Account timeline
    result = run(cli.timelines.timeline, "--account", friend_user.username)
    assert result.exit_code == 0
    assert status1.id not in result.stdout
    assert status2.id not in result.stdout
    assert status3.id in result.stdout

    result = run(cli.timelines.timeline, "--account", other_user.username)
    assert result.exit_code == 0
    assert status1.id not in result.stdout
    assert status2.id in result.stdout
    assert status3.id not in result.stdout


def test_empty_timeline(app, run_as):
    user = register_account(app)
    result = run_as(user, cli.timelines.timeline)
    assert result.exit_code == 0
    assert result.stdout.strip() == "─" * 80


def test_timeline_cant_combine_timelines(run):
    result = run(cli.timelines.timeline, "--tag", "foo", "--account", "bar")
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: Only one of --public, --tag, --account, or --list can be used at one time."


def test_timeline_local_needs_public_or_tag(run):
    result = run(cli.timelines.timeline, "--local")
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: The --local option is only valid alongside --public or --tag."


def test_timeline_instance_needs_public_or_tag(run):
    result = run(cli.timelines.timeline, "--instance", TOOT_TEST_BASE_URL)
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: The --instance option is only valid alongside --public or --tag."


def test_bookmarks(app, user, run):
    status1 = _post_status(app, user)
    status2 = _post_status(app, user)

    api.bookmark(app, user, status1.id)
    api.bookmark(app, user, status2.id)

    result = run(cli.timelines.bookmarks)
    assert result.exit_code == 0
    assert status1.id in result.stdout
    assert status2.id in result.stdout
    assert result.stdout.find(status1.id) > result.stdout.find(status2.id)


    result = run(cli.timelines.bookmarks, "--reverse")
    assert result.exit_code == 0
    assert status1.id in result.stdout
    assert status2.id in result.stdout
    assert result.stdout.find(status1.id) < result.stdout.find(status2.id)


def test_notifications(app, user, other_user, run):
    result = run(cli.timelines.notifications)
    assert result.exit_code == 0
    assert result.stdout.strip() == "You have no notifications"

    text = f"Paging doctor @{user.username}"
    status = _post_status(app, other_user, text)

    def test_notifications():
        result = run(cli.timelines.notifications)
        assert result.exit_code == 0
        assert f"@{other_user.username} mentioned you" in result.stdout
        assert status.id in result.stdout
        assert text in result.stdout
    run_with_retries(test_notifications)

    result = run(cli.timelines.notifications, "--mentions")
    assert result.exit_code == 0
    assert f"@{other_user.username} mentioned you" in result.stdout
    assert status.id in result.stdout
    assert text in result.stdout


def test_notifications_follow(app, user, friend_user, run_as):
    result = run_as(friend_user, cli.timelines.notifications)
    assert result.exit_code == 0
    assert f"@{user.username} now follows you" in result.stdout

    result = run_as(friend_user, cli.timelines.notifications, "--mentions")
    assert result.exit_code == 0
    assert "now follows you" not in result.stdout


def _post_status(app, user, text=None) -> Status:
    text = text or str(uuid4())
    response = api.post_status(app, user, text)
    return from_dict(Status, response.json())



from toot import api, config, User, App
from toot.entities import from_dict, Instance
from toot.exceptions import ApiError, ConsoleError
from urllib.parse import urlparse


def find_instance(base_url: str) -> Instance:
    try:
        instance = api.get_instance(base_url).json()
        return from_dict(Instance, instance)
    except Exception:
        raise ConsoleError(f"Instance not found at {base_url}")


def register_app(domain: str, base_url: str) -> App:
    try:
        response = api.create_app(base_url)
    except ApiError:
        raise ConsoleError("Registration failed.")

    app = App(domain, base_url, response['client_id'], response['client_secret'])
    config.save_app(app)

    return app


def get_or_create_app(base_url: str) -> App:
    instance = find_instance(base_url)
    domain = _get_instance_domain(instance)
    return config.load_app(domain) or register_app(domain, base_url)


def create_user(app: App, access_token: str) -> User:
    # Username is not yet known at this point, so fetch it from Mastodon
    user = User(app.instance, None, access_token)
    creds = api.verify_credentials(app, user).json()

    user = User(app.instance, creds["username"], access_token)
    config.save_user(user, activate=True)

    return user


def login_username_password(app: App, email: str, password: str) -> User:
    try:
        response = api.login(app, email, password)
    except Exception:
        raise ConsoleError("Login failed")

    return create_user(app, response["access_token"])


def login_auth_code(app: App, authorization_code: str) -> User:
    try:
        response = api.request_access_token(app, authorization_code)
    except Exception:
        raise ConsoleError("Login failed")

    return create_user(app, response["access_token"])


def _get_instance_domain(instance: Instance) -> str:
    """Extracts the instance domain name.

    Pleroma and its forks return an actual URI here, rather than a domain name
    like Mastodon. This is contrary to the spec.¯ in that case, parse out the
    domain and return it.

    TODO: when updating to v2 instance endpoint, this field has been renamed to
    `domain`
    """
    if instance.uri.startswith("http"):
        return urlparse(instance.uri).netloc
    return instance.uri



from uuid import uuid4
from tests.integration.conftest import TRUMPET
from toot import api, cli
from toot.entities import Account, from_dict
from toot.utils import get_text


def test_update_account_no_options(run):
    result = run(cli.accounts.update_account)
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: Please specify at least one option to update the account"


def test_update_account_display_name(run, app, user):
    name = str(uuid4())[:10]

    result = run(cli.accounts.update_account, "--display-name", name)
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account updated"

    account = api.verify_credentials(app, user).json()
    assert account["display_name"] == name


def test_update_account_json(run_json, app, user):
    name = str(uuid4())[:10]
    out = run_json(cli.accounts.update_account, "--display-name", name, "--json")
    account = from_dict(Account, out)
    assert account.acct == user.username
    assert account.display_name == name


def test_update_account_note(run, app, user):
    note = ("It's 106 miles to Chicago, we got a full tank of gas, half a pack "
           "of cigarettes, it's dark... and we're wearing sunglasses.")

    result = run(cli.accounts.update_account, "--note", note)
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account updated"

    account = api.verify_credentials(app, user).json()
    assert get_text(account["note"]) == note


def test_update_account_language(run, app, user):
    result = run(cli.accounts.update_account, "--language", "hr")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account updated"

    account = api.verify_credentials(app, user).json()
    assert account["source"]["language"] == "hr"


def test_update_account_privacy(run, app, user):
    result = run(cli.accounts.update_account, "--privacy", "private")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account updated"

    account = api.verify_credentials(app, user).json()
    assert account["source"]["privacy"] == "private"


def test_update_account_avatar(run, app, user):
    account = api.verify_credentials(app, user).json()
    old_value = account["avatar"]

    result = run(cli.accounts.update_account, "--avatar", TRUMPET)
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account updated"

    account = api.verify_credentials(app, user).json()
    assert account["avatar"] != old_value


def test_update_account_header(run, app, user):
    account = api.verify_credentials(app, user).json()
    old_value = account["header"]

    result = run(cli.accounts.update_account, "--header", TRUMPET)
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account updated"

    account = api.verify_credentials(app, user).json()
    assert account["header"] != old_value


def test_update_account_locked(run, app, user):
    result = run(cli.accounts.update_account, "--locked")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account updated"

    account = api.verify_credentials(app, user).json()
    assert account["locked"] is True

    result = run(cli.accounts.update_account, "--no-locked")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account updated"

    account = api.verify_credentials(app, user).json()
    assert account["locked"] is False


def test_update_account_bot(run, app, user):
    result = run(cli.accounts.update_account, "--bot")

    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account updated"

    account = api.verify_credentials(app, user).json()
    assert account["bot"] is True

    result = run(cli.accounts.update_account, "--no-bot")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account updated"

    account = api.verify_credentials(app, user).json()
    assert account["bot"] is False


def test_update_account_discoverable(run, app, user):
    result = run(cli.accounts.update_account, "--discoverable")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account updated"

    account = api.verify_credentials(app, user).json()
    assert account["discoverable"] is True

    result = run(cli.accounts.update_account, "--no-discoverable")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account updated"

    account = api.verify_credentials(app, user).json()
    assert account["discoverable"] is False


def test_update_account_sensitive(run, app, user):
    result = run(cli.accounts.update_account, "--sensitive")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account updated"

    account = api.verify_credentials(app, user).json()
    assert account["source"]["sensitive"] is True

    result = run(cli.accounts.update_account, "--no-sensitive")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account updated"

    account = api.verify_credentials(app, user).json()
    assert account["source"]["sensitive"] is False



from typing import Any, Dict
from unittest import mock
from unittest.mock import MagicMock

from toot import User, cli
from tests.integration.conftest import Run

# TODO: figure out how to test login


EMPTY_CONFIG: Dict[Any, Any] = {
    "apps": {},
    "users": {},
    "active_user": None
}

SAMPLE_CONFIG = {
    "active_user": "frank@foo.social",
    "apps": {
        "foo.social": {
            "base_url": "http://foo.social",
            "client_id": "123",
            "client_secret": "123",
            "instance": "foo.social"
        },
        "bar.social": {
            "base_url": "http://bar.social",
            "client_id": "123",
            "client_secret": "123",
            "instance": "bar.social"
        },
    },
    "users": {
        "frank@foo.social": {
            "access_token": "123",
            "instance": "foo.social",
            "username": "frank"
        },
        "frank@bar.social": {
            "access_token": "123",
            "instance": "bar.social",
            "username": "frank"
        },
    }
}


def test_env(run: Run):
    result = run(cli.auth.env)
    assert result.exit_code == 0
    assert "toot" in result.stdout
    assert "Python" in result.stdout


@mock.patch("toot.config.load_config")
def test_auth_empty(load_config: MagicMock, run: Run):
    load_config.return_value = EMPTY_CONFIG
    result = run(cli.auth.auth)
    assert result.exit_code == 0
    assert result.stdout.strip() == "You are not logged in to any accounts"


@mock.patch("toot.config.load_config")
def test_auth_full(load_config: MagicMock, run: Run):
    load_config.return_value = SAMPLE_CONFIG
    result = run(cli.auth.auth)
    assert result.exit_code == 0
    assert result.stdout.strip().startswith("Authenticated accounts:")
    assert "frank@foo.social" in result.stdout
    assert "frank@bar.social" in result.stdout


# Saving config is mocked so we don't mess up our local config
# TODO: could this be implemented using an auto-use fixture so we have it always
# mocked?
@mock.patch("toot.config.load_app")
@mock.patch("toot.config.save_app")
@mock.patch("toot.config.save_user")
def test_login_cli(
    save_user: MagicMock,
    save_app: MagicMock,
    load_app: MagicMock,
    user: User,
    run: Run,
):
    load_app.return_value = None

    result = run(
        cli.auth.login_cli,
        "--instance", "http://localhost:3000",
        "--email", f"{user.username}@example.com",
        "--password", "password",
    )
    assert result.exit_code == 0
    assert "✓ Successfully logged in." in result.stdout

    save_app.assert_called_once()
    (app,) = save_app.call_args.args
    assert app.instance == "localhost:3000"
    assert app.base_url == "http://localhost:3000"
    assert app.client_id
    assert app.client_secret

    save_user.assert_called_once()
    (new_user,) = save_user.call_args.args
    assert new_user.instance == "localhost:3000"
    assert new_user.username == user.username
    # access token will be different since this is a new login
    assert new_user.access_token and new_user.access_token != user.access_token
    assert save_user.call_args.kwargs == {"activate": True}


@mock.patch("toot.config.load_app")
@mock.patch("toot.config.save_app")
@mock.patch("toot.config.save_user")
def test_login_cli_wrong_password(
    save_user: MagicMock,
    save_app: MagicMock,
    load_app: MagicMock,
    user: User,
    run: Run,
):
    load_app.return_value = None

    result = run(
        cli.auth.login_cli,
        "--instance", "http://localhost:3000",
        "--email", f"{user.username}@example.com",
        "--password", "wrong password",
    )
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: Login failed"

    save_app.assert_called_once()
    (app,) = save_app.call_args.args
    assert app.instance == "localhost:3000"
    assert app.base_url == "http://localhost:3000"
    assert app.client_id
    assert app.client_secret

    save_user.assert_not_called()


@mock.patch("toot.config.load_config")
@mock.patch("toot.config.delete_user")
def test_logout(delete_user: MagicMock, load_config: MagicMock, run: Run):
    load_config.return_value = SAMPLE_CONFIG

    result = run(cli.auth.logout, "frank@foo.social")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account frank@foo.social logged out"
    delete_user.assert_called_once_with(User("foo.social", "frank", "123"))


@mock.patch("toot.config.load_config")
def test_logout_not_logged_in(load_config: MagicMock, run: Run):
    load_config.return_value = EMPTY_CONFIG

    result = run(cli.auth.logout)
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: You're not logged into any accounts"


@mock.patch("toot.config.load_config")
def test_logout_account_not_specified(load_config: MagicMock, run: Run):
    load_config.return_value = SAMPLE_CONFIG

    result = run(cli.auth.logout)
    assert result.exit_code == 1
    assert result.stderr.startswith("Error: Specify account to log out")


@mock.patch("toot.config.load_config")
def test_logout_account_does_not_exist(load_config: MagicMock, run: Run):
    load_config.return_value = SAMPLE_CONFIG

    result = run(cli.auth.logout, "banana")
    assert result.exit_code == 1
    assert result.stderr.startswith("Error: Account not found")


@mock.patch("toot.config.load_config")
@mock.patch("toot.config.activate_user")
def test_activate(activate_user: MagicMock, load_config: MagicMock, run: Run):
    load_config.return_value = SAMPLE_CONFIG

    result = run(cli.auth.activate, "frank@foo.social")
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Account frank@foo.social activated"
    activate_user.assert_called_once_with(User("foo.social", "frank", "123"))


@mock.patch("toot.config.load_config")
def test_activate_not_logged_in(load_config: MagicMock, run: Run):
    load_config.return_value = EMPTY_CONFIG

    result = run(cli.auth.activate)
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: You're not logged into any accounts"


@mock.patch("toot.config.load_config")
def test_activate_account_not_given(load_config: MagicMock, run: Run):
    load_config.return_value = SAMPLE_CONFIG

    result = run(cli.auth.activate)
    assert result.exit_code == 1
    assert result.stderr.startswith("Error: Specify account to activate")


@mock.patch("toot.config.load_config")
def test_activate_invalid_Account(load_config: MagicMock, run: Run):
    load_config.return_value = SAMPLE_CONFIG

    result = run(cli.auth.activate, "banana")
    assert result.exit_code == 1
    assert result.stderr.startswith("Error: Account not found")



from uuid import uuid4
from toot import cli

from tests.integration.conftest import register_account


def test_lists_empty(run):
    result = run(cli.lists.list)
    assert result.exit_code == 0
    assert result.stdout.strip() == "You have no lists defined."


def test_lists_empty_json(run_json):
    lists = run_json(cli.lists.list, "--json")
    assert lists == []


def test_list_create_delete(run):
    result = run(cli.lists.create, "banana")
    assert result.exit_code == 0
    assert result.stdout.strip() == '✓ List "banana" created.'

    result = run(cli.lists.list)
    assert result.exit_code == 0
    assert "banana" in result.stdout

    result = run(cli.lists.create, "mango")
    assert result.exit_code == 0
    assert result.stdout.strip() == '✓ List "mango" created.'

    result = run(cli.lists.list)
    assert result.exit_code == 0
    assert "banana" in result.stdout
    assert "mango" in result.stdout

    result = run(cli.lists.delete, "banana")
    assert result.exit_code == 0
    assert result.stdout.strip() == '✓ List "banana" deleted.'

    result = run(cli.lists.list)
    assert result.exit_code == 0
    assert "banana" not in result.stdout
    assert "mango" in result.stdout

    result = run(cli.lists.delete, "mango")
    assert result.exit_code == 0
    assert result.stdout.strip() == '✓ List "mango" deleted.'

    result = run(cli.lists.list)
    assert result.exit_code == 0
    assert result.stdout.strip() == "You have no lists defined."

    result = run(cli.lists.delete, "mango")
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: List not found"


def test_list_create_delete_json(run, run_json):
    result = run_json(cli.lists.list, "--json")
    assert result == []

    list = run_json(cli.lists.create, "banana", "--json")
    assert list["title"] == "banana"

    [list] = run_json(cli.lists.list, "--json")
    assert list["title"] == "banana"

    list = run_json(cli.lists.create, "mango", "--json")
    assert list["title"] == "mango"

    lists = run_json(cli.lists.list, "--json")
    [list1, list2] = sorted(lists, key=lambda l: l["title"])
    assert list1["title"] == "banana"
    assert list2["title"] == "mango"

    result = run_json(cli.lists.delete, "banana", "--json")
    assert result == {}

    [list] = run_json(cli.lists.list, "--json")
    assert list["title"] == "mango"

    result = run_json(cli.lists.delete, "mango", "--json")
    assert result == {}

    result = run_json(cli.lists.list, "--json")
    assert result == []

    result = run(cli.lists.delete, "mango", "--json")
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: List not found"


def test_list_add_remove(run, app):
    list_name = str(uuid4())
    acc = register_account(app)
    run(cli.lists.create, list_name)

    result = run(cli.lists.add, list_name, acc.username)
    assert result.exit_code == 1
    assert result.stderr.strip() == f"Error: You must follow @{acc.username} before adding this account to a list."

    run(cli.accounts.follow, acc.username)

    result = run(cli.lists.add, list_name, acc.username)
    assert result.exit_code == 0
    assert result.stdout.strip() == f'✓ Added account "{acc.username}"'

    result = run(cli.lists.accounts, list_name)
    assert result.exit_code == 0
    assert acc.username in result.stdout

    # Account doesn't exist
    result = run(cli.lists.add, list_name, "does_not_exist")
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: Account not found"

    # List doesn't exist
    result = run(cli.lists.add, "does_not_exist", acc.username)
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: List not found"

    result = run(cli.lists.remove, list_name, acc.username)
    assert result.exit_code == 0
    assert result.stdout.strip() == f'✓ Removed account "{acc.username}"'

    result = run(cli.lists.accounts, list_name)
    assert result.exit_code == 0
    assert result.stdout.strip() == "This list has no accounts."


def test_list_add_remove_json(run, run_json, app):
    list_name = str(uuid4())
    acc = register_account(app)
    run(cli.lists.create, list_name)

    result = run(cli.lists.add, list_name, acc.username, "--json")
    assert result.exit_code == 1
    assert result.stderr.strip() == f"Error: You must follow @{acc.username} before adding this account to a list."

    run(cli.accounts.follow, acc.username)

    result = run_json(cli.lists.add, list_name, acc.username, "--json")
    assert result == {}

    [account] = run_json(cli.lists.accounts, list_name, "--json")
    assert account["username"] == acc.username

    # Account doesn't exist
    result = run(cli.lists.add, list_name, "does_not_exist", "--json")
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: Account not found"

    # List doesn't exist
    result = run(cli.lists.add, "does_not_exist", acc.username, "--json")
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: List not found"

    result = run_json(cli.lists.remove, list_name, acc.username, "--json")
    assert result == {}

    result = run_json(cli.lists.accounts, list_name, "--json")
    assert result == []



# Languages mapped by their ISO 639-1 code
# https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes
LANGUAGES = {
    "ab": "Abkhazian",
    "aa": "Afar",
    "af": "Afrikaans",
    "ak": "Akan",
    "sq": "Albanian",
    "am": "Amharic",
    "ar": "Arabic",
    "an": "Aragonese",
    "hy": "Armenian",
    "as": "Assamese",
    "av": "Avaric",
    "ae": "Avestan",
    "ay": "Aymara",
    "az": "Azerbaijani",
    "bm": "Bambara",
    "ba": "Bashkir",
    "eu": "Basque",
    "be": "Belarusian",
    "bn": "Bengali",
    "bi": "Bislama",
    "bs": "Bosnian",
    "br": "Breton",
    "bg": "Bulgarian",
    "my": "Burmese",
    "ca": "Catalan",
    "ch": "Chamorro",
    "ce": "Chechen",
    "ny": "Chichewa",
    "zh": "Chinese",
    "cu": "Old Slavonic",
    "cv": "Chuvash",
    "kw": "Cornish",
    "co": "Corsican",
    "cr": "Cree",
    "hr": "Croatian",
    "cs": "Czech",
    "da": "Danish",
    "dv": "Divehi",
    "nl": "Dutch",
    "en": "English",
    "eo": "Estonian",
    "ee": "Ewe",
    "fo": "Faroese",
    "fj": "Fijian",
    "fi": "Finnish",
    "fr": "French",
    "fy": "Western Frisian",
    "ff": "Fulah",
    "gd": "Gaelic",
    "gl": "Galician",
    "lg": "Ganda",
    "ka": "Georgian",
    "de": "German",
    "el": "Greek",
    "kl": "Kalaallisut",
    "gn": "Guarani",
    "gu": "Gujarati",
    "ht": "Haitian",
    "ha": "Hausa",
    "he": "Hebrew",
    "hz": "Herero",
    "hi": "Hiri Motu",
    "hu": "Hungarian",
    "is": "Icelandic",
    "io": "Ido",
    "ig": "Igbo",
    "id": "Indonesian",
    "ia": "Inupiaq",
    "ga": "Irish",
    "it": "Italian",
    "ja": "Japanese",
    "jv": "Javanese",
    "kn": "Kannada",
    "kr": "Kanuri",
    "ks": "Kashmiri",
    "kk": "Kazakh",
    "km": "Central Khmer",
    "ki": "Kikuyu",
    "rw": "Kirghiz",
    "kv": "Komi",
    "kg": "Kongo",
    "ko": "Korean",
    "kj": "Kuanyama",
    "ku": "Kurdish",
    "lo": "Lao",
    "la": "Latvian",
    "li": "Limburgan",
    "ln": "Lingala",
    "lt": "Lithuanian",
    "lu": "Luba-Katanga",
    "lb": "Luxembourgish",
    "mk": "Macedonian",
    "mg": "Malagasy",
    "ms": "Malay",
    "ml": "Malayalam",
    "mt": "Maltese",
    "gv": "Manx",
    "mi": "Maori",
    "mr": "Marathi",
    "mh": "Marshallese",
    "mn": "Mongolian",
    "na": "Nauru",
    "nv": "Navajo",
    "nd": "North Ndebele",
    "nr": "South Ndebele",
    "ng": "Nepali",
    "no": "Norwegian",
    "nb": "Norwegian Bokmål",
    "nn": "Norwegian Nynorsk",
    "ii": "Sichuan Yi",
    "oc": "Occitan",
    "oj": "Ojibwa",
    "or": "Oriya",
    "om": "Oromo",
    "os": "Ossetian",
    "pi": "Pali",
    "ps": "Pashto",
    "fa": "Persian",
    "pl": "Polish",
    "pt": "Portuguese",
    "pa": "Punjabi",
    "qu": "Quechua",
    "ro": "Romanian",
    "rm": "Romansh",
    "rn": "Rundi",
    "ru": "Russian",
    "se": "Samoan",
    "sg": "Sango",
    "sa": "Sardinian",
    "sr": "Serbian",
    "sn": "Shona",
    "sd": "Sindhi",
    "si": "Sinhala",
    "sk": "Slovak",
    "sl": "Slovenian",
    "so": "Somali",
    "st": "Southern Sotho",
    "es": "Spanish",
    "su": "Sundanese",
    "sw": "Swahili",
    "ss": "Swati",
    "sv": "Swedish",
    "tl": "Tagalog",
    "ty": "Tahitian",
    "tg": "Tajik",
    "ta": "Tamil",
    "tt": "Tatar",
    "te": "Telugu",
    "th": "Thai",
    "bo": "Tibetan",
    "ti": "Tigrinya",
    "to": "Tonga",
    "ts": "Tsonga",
    "tn": "Tswana",
    "tr": "Turkish",
    "tk": "Turkmen",
    "tw": "Uighur",
    "uk": "Ukrainian",
    "ur": "Uzbek",
    "ve": "Venda",
    "vi": "Vietnamese",
    "vo": "Walloon",
    "cy": "Welsh",
    "wo": "Wolof",
    "xh": "Xhosa",
    "yi": "Yiddish",
    "yo": "Yoruba",
    "za": "Zhuang",
    "zu": "Zulu",
}


def language_name(code):
    return LANGUAGES.get(code, code)



from collections import namedtuple

from toot.utils.datetime import parse_datetime

Author = namedtuple("Author", ["account", "display_name", "username"])


class Status:
    """
    A wrapper around the Status entity data fetched from Mastodon.

    https://github.com/tootsuite/documentation/blob/master/Using-the-API/API.md#status

    Attributes
    ----------
    reblog : Status or None
        The reblogged status if it exists.

    original : Status
        If a reblog, the reblogged status, otherwise self.
    """

    def __init__(self, data, is_mine, default_instance):
        """
        Parameters
        ----------
        data : dict
            Status data as received from Mastodon.
            https://docs.joinmastodon.org/api/entities/#status

        is_mine : bool
            Whether the status was created by the logged in user.

        default_instance : str
            The domain of the instance into which the user is logged in. Used to
            create fully qualified account names for users on the same instance.
            Mastodon only populates the name, not the domain.
        """

        self.data = data
        self.is_mine = is_mine
        self.default_instance = default_instance

        # This can be toggled by the user
        self.show_sensitive = False

        # Set when status is translated
        self.show_translation = False
        self.translation = None
        self.translated_from = None

        # TODO: clean up
        self.id = self.data["id"]
        self.account = self._get_account()
        self.created_at = parse_datetime(data["created_at"])
        if data["edited_at"]:
            self.edited_at = parse_datetime(data["edited_at"])
        else:
            self.edited_at = None
        self.author = self._get_author()
        self.favourited = data.get("favourited", False)
        self.reblogged = data.get("reblogged", False)
        self.bookmarked = data.get("bookmarked", False)
        self.in_reply_to = data.get("in_reply_to_id")
        self.url = data.get("url")
        self.mentions = data.get("mentions")
        self.reblog = self._get_reblog()
        self.visibility = data.get("visibility")

    @property
    def original(self):
        return self.reblog or self

    def _get_reblog(self):
        reblog = self.data.get("reblog")
        if not reblog:
            return None

        reblog_is_mine = self.is_mine and (
            self.data["account"]["acct"] == reblog["account"]["acct"]
        )
        return Status(reblog, reblog_is_mine, self.default_instance)

    def _get_author(self):
        acct = self.data['account']['acct']
        acct = acct if "@" in acct else "{}@{}".format(acct, self.default_instance)
        return Author(acct, self.data['account']['display_name'], self.data['account']['username'])

    def _get_account(self):
        acct = self.data['account']['acct']
        return acct if "@" in acct else "{}@{}".format(acct, self.default_instance)

    def __repr__(self):
        return "<Status id={} account={}>".format(self.id, self.account)



import json
import re

from tests.integration.conftest import TOOT_TEST_BASE_URL
from toot import api, cli
from toot.entities import Account, Status, from_dict, from_dict_list
from uuid import uuid4


def test_instance_default(app, run):
    result = run(cli.read.instance)
    assert result.exit_code == 0

    assert "Mastodon" in result.stdout
    assert app.instance in result.stdout
    assert "running Mastodon" in result.stdout


def test_instance_with_url(app, run):
    result = run(cli.read.instance, TOOT_TEST_BASE_URL)
    assert result.exit_code == 0

    assert "Mastodon" in result.stdout
    assert app.instance in result.stdout
    assert "running Mastodon" in result.stdout


def test_instance_json(app, run):
    result = run(cli.read.instance, "--json")
    assert result.exit_code == 0

    data = json.loads(result.stdout)
    assert data["title"] is not None
    assert data["description"] is not None
    assert data["version"] is not None


def test_instance_anon(app, run_anon, base_url):
    result = run_anon(cli.read.instance, base_url)
    assert result.exit_code == 0

    assert "Mastodon" in result.stdout
    assert app.instance in result.stdout
    assert "running Mastodon" in result.stdout

    # Need to specify the instance name when running anon
    result = run_anon(cli.read.instance)
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: INSTANCE argument not given and not logged in"


def test_whoami(user, run):
    result = run(cli.read.whoami)
    assert result.exit_code == 0
    assert f"@{user.username}" in result.stdout


def test_whoami_json(user, run):
    result = run(cli.read.whoami, "--json")
    assert result.exit_code == 0

    data = json.loads(result.stdout)
    account = from_dict(Account, data)
    assert account.username == user.username
    assert account.acct == user.username


def test_whois(app, friend, run):
    variants = [
        friend.username,
        f"@{friend.username}",
        f"{friend.username}@{app.instance}",
        f"@{friend.username}@{app.instance}",
    ]

    for username in variants:
        result = run(cli.read.whois, username)
        assert result.exit_code == 0
        assert f"@{friend.username}" in result.stdout


def test_whois_json(app, friend, run):
    result = run(cli.read.whois, friend.username, "--json")
    assert result.exit_code == 0

    data = json.loads(result.stdout)
    account = from_dict(Account, data)
    assert account.username == friend.username
    assert account.acct == friend.username


def test_search_account(friend, run):
    result = run(cli.read.search, friend.username)
    assert result.exit_code == 0
    assert result.stdout.strip() == f"Accounts:\n* @{friend.username}"


def test_search_account_json(friend, run):
    result = run(cli.read.search, friend.username, "--json")
    assert result.exit_code == 0

    data = json.loads(result.stdout)
    [account] = from_dict_list(Account, data["accounts"])
    assert account.acct == friend.username


def test_search_hashtag(app, user, run):
    api.post_status(app, user, "#hashtag_x")
    api.post_status(app, user, "#hashtag_y")
    api.post_status(app, user, "#hashtag_z")

    result = run(cli.read.search, "#hashtag")
    assert result.exit_code == 0
    assert result.stdout.strip() == "Hashtags:\n#hashtag_x, #hashtag_y, #hashtag_z"


def test_search_hashtag_json(app, user, run):
    api.post_status(app, user, "#hashtag_x")
    api.post_status(app, user, "#hashtag_y")
    api.post_status(app, user, "#hashtag_z")

    result = run(cli.read.search, "#hashtag", "--json")
    assert result.exit_code == 0

    data = json.loads(result.stdout)
    [h1, h2, h3] = sorted(data["hashtags"], key=lambda h: h["name"])

    assert h1["name"] == "hashtag_x"
    assert h2["name"] == "hashtag_y"
    assert h3["name"] == "hashtag_z"


def test_status(app, user, run):
    uuid = str(uuid4())
    status_id = api.post_status(app, user, uuid).json()["id"]

    result = run(cli.read.status, status_id)
    assert result.exit_code == 0

    out = result.stdout.strip()
    assert uuid in out
    assert user.username in out
    assert status_id in out


def test_status_json(app, user, run):
    uuid = str(uuid4())
    status_id = api.post_status(app, user, uuid).json()["id"]

    result = run(cli.read.status, status_id, "--json")
    assert result.exit_code == 0

    status = from_dict(Status, json.loads(result.stdout))
    assert status.id == status_id
    assert status.account.acct == user.username
    assert uuid in status.content


def test_thread(app, user, run):
    uuid1 = str(uuid4())
    uuid2 = str(uuid4())
    uuid3 = str(uuid4())

    s1 = api.post_status(app, user, uuid1).json()
    s2 = api.post_status(app, user, uuid2, in_reply_to_id=s1["id"]).json()
    s3 = api.post_status(app, user, uuid3, in_reply_to_id=s2["id"]).json()

    for status in [s1, s2, s3]:
        result = run(cli.read.thread, status["id"])
        assert result.exit_code == 0

        bits = re.split(r"─+", result.stdout.strip())
        bits = [b for b in bits if b]

        assert len(bits) == 3

        assert s1["id"] in bits[0]
        assert s2["id"] in bits[1]
        assert s3["id"] in bits[2]

        assert uuid1 in bits[0]
        assert uuid2 in bits[1]
        assert uuid3 in bits[2]


def test_thread_json(app, user, run):
    uuid1 = str(uuid4())
    uuid2 = str(uuid4())
    uuid3 = str(uuid4())

    s1 = api.post_status(app, user, uuid1).json()
    s2 = api.post_status(app, user, uuid2, in_reply_to_id=s1["id"]).json()
    s3 = api.post_status(app, user, uuid3, in_reply_to_id=s2["id"]).json()

    result = run(cli.read.thread, s2["id"], "--json")
    assert result.exit_code == 0

    result = json.loads(result.stdout)
    [ancestor] = [from_dict(Status, s) for s in result["ancestors"]]
    [descendent] = [from_dict(Status, s) for s in result["descendants"]]

    assert ancestor.id == s1["id"]
    assert descendent.id == s3["id"]



"""
This module contains integration tests meant to run against a test Mastodon instance.

You can set up a test instance locally by following this guide:
https://docs.joinmastodon.org/dev/setup/

To enable integration tests, export the following environment variables to match
your test server and database:

```
export TOOT_TEST_BASE_URL="localhost:3000"
```
"""

import json
import os
import pytest
import re
import typing as t
import uuid

from click.testing import CliRunner, Result
from pathlib import Path
from toot import api, App, User
from toot.cli import Context, TootObj


def pytest_configure(config):
    import toot.settings
    toot.settings.DISABLE_SETTINGS = True


# Type alias for run commands
Run = t.Callable[..., Result]

# Mastodon database name, used to confirm user registration without having to click the link
TOOT_TEST_BASE_URL = os.getenv("TOOT_TEST_BASE_URL")

# Toot logo used for testing image upload
TRUMPET = str(Path(__file__).parent.parent.parent / "trumpet.png")

ASSETS_DIR = str(Path(__file__).parent.parent / "assets")


def create_app(base_url):
    instance = api.get_instance(base_url).json()
    response = api.create_app(base_url)
    return App(instance["uri"], base_url, response["client_id"], response["client_secret"])


def register_account(app: App):
    username = str(uuid.uuid4())[-10:]
    email = f"{username}@example.com"

    response = api.register_account(app, username, email, "password", "en")
    return User(app.instance, username, response["access_token"])


# ------------------------------------------------------------------------------
# Fixtures
# ------------------------------------------------------------------------------


# Host name of a test instance to run integration tests against
# DO NOT USE PUBLIC INSTANCES!!!
@pytest.fixture(scope="session")
def base_url():
    if not TOOT_TEST_BASE_URL:
        pytest.skip("Skipping integration tests, TOOT_TEST_BASE_URL not set")

    return TOOT_TEST_BASE_URL


@pytest.fixture(scope="session")
def app(base_url):
    return create_app(base_url)


@pytest.fixture(scope="session")
def user(app):
    return register_account(app)


@pytest.fixture(scope="session")
def friend(app):
    return register_account(app)


@pytest.fixture(scope="session")
def user_id(app, user):
    return api.find_account(app, user, user.username)["id"]


@pytest.fixture(scope="session")
def friend_id(app, user, friend):
    return api.find_account(app, user, friend.username)["id"]


@pytest.fixture(scope="session", autouse=True)
def testing_env():
    os.environ["TOOT_TESTING"] = "true"


@pytest.fixture(scope="session")
def runner():
    return CliRunner(mix_stderr=False)


@pytest.fixture
def run(app, user, runner):
    def _run(command, *params, input=None) -> Result:
        obj = TootObj(test_ctx=Context(app, user))
        return runner.invoke(command, params, obj=obj, input=input)
    return _run


@pytest.fixture
def run_as(app, runner):
    def _run_as(user, command, *params, input=None) -> Result:
        obj = TootObj(test_ctx=Context(app, user))
        return runner.invoke(command, params, obj=obj, input=input)
    return _run_as


@pytest.fixture
def run_json(app, user, runner):
    def _run_json(command, *params):
        obj = TootObj(test_ctx=Context(app, user))
        result = runner.invoke(command, params, obj=obj)
        assert result.exit_code == 0
        return json.loads(result.stdout)
    return _run_json


@pytest.fixture
def run_anon(runner):
    def _run(command, *params) -> Result:
        obj = TootObj(test_ctx=Context(None, None))
        return runner.invoke(command, params, obj=obj)
    return _run


# ------------------------------------------------------------------------------
# Utils
# ------------------------------------------------------------------------------


def posted_status_id(out):
    pattern = re.compile(r"Toot posted: http://([^/]+)/([^/]+)/(.+)")
    match = re.search(pattern, out)
    assert match

    _, _, status_id = match.groups()

    return status_id



import json
import pytest

from tests.utils import run_with_retries
from toot import api, cli
from toot.exceptions import NotFoundError


def test_delete(app, user, run):
    status = api.post_status(app, user, "foo").json()

    result = run(cli.statuses.delete, status["id"])
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Status deleted"

    with pytest.raises(NotFoundError):
        api.fetch_status(app, user, status["id"])


def test_delete_json(app, user, run):
    status = api.post_status(app, user, "foo").json()

    result = run(cli.statuses.delete, status["id"], "--json")
    assert result.exit_code == 0

    out = result.stdout
    result = json.loads(out)
    assert result["id"] == status["id"]

    with pytest.raises(NotFoundError):
        api.fetch_status(app, user, status["id"])


def test_favourite(app, user, run):
    status = api.post_status(app, user, "foo").json()
    assert not status["favourited"]

    result = run(cli.statuses.favourite, status["id"])
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Status favourited"

    status = api.fetch_status(app, user, status["id"]).json()
    assert status["favourited"]

    result = run(cli.statuses.unfavourite, status["id"])
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Status unfavourited"

    def test_favourited():
        nonlocal status
        status = api.fetch_status(app, user, status["id"]).json()
        assert not status["favourited"]
    run_with_retries(test_favourited)


def test_favourite_json(app, user, run):
    status = api.post_status(app, user, "foo").json()
    assert not status["favourited"]

    result = run(cli.statuses.favourite, status["id"], "--json")
    assert result.exit_code == 0

    result = json.loads(result.stdout)
    assert result["id"] == status["id"]
    assert result["favourited"] is True

    result = run(cli.statuses.unfavourite, status["id"], "--json")
    assert result.exit_code == 0

    result = json.loads(result.stdout)
    assert result["id"] == status["id"]
    assert result["favourited"] is False


def test_reblog(app, user, run):
    status = api.post_status(app, user, "foo").json()
    assert not status["reblogged"]

    result = run(cli.statuses.reblogged_by, status["id"])
    assert result.exit_code == 0
    assert result.stdout.strip() == "This status is not reblogged by anyone"

    result = run(cli.statuses.reblog, status["id"])
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Status reblogged"

    status = api.fetch_status(app, user, status["id"]).json()
    assert status["reblogged"]

    result = run(cli.statuses.reblogged_by, status["id"])
    assert result.exit_code == 0
    assert user.username in result.stdout

    result = run(cli.statuses.unreblog, status["id"])
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Status unreblogged"

    status = api.fetch_status(app, user, status["id"]).json()
    assert not status["reblogged"]


def test_reblog_json(app, user, run):
    status = api.post_status(app, user, "foo").json()
    assert not status["reblogged"]

    result = run(cli.statuses.reblog, status["id"], "--json")
    assert result.exit_code == 0

    result = json.loads(result.stdout)
    assert result["reblogged"] is True
    assert result["reblog"]["id"] == status["id"]

    result = run(cli.statuses.reblogged_by, status["id"], "--json")
    assert result.exit_code == 0

    [reblog] = json.loads(result.stdout)
    assert reblog["acct"] == user.username

    result = run(cli.statuses.unreblog, status["id"], "--json")
    assert result.exit_code == 0

    result = json.loads(result.stdout)
    assert result["reblogged"] is False
    assert result["reblog"] is None


def test_pin(app, user, run):
    status = api.post_status(app, user, "foo").json()
    assert not status["pinned"]

    result = run(cli.statuses.pin, status["id"])
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Status pinned"

    status = api.fetch_status(app, user, status["id"]).json()
    assert status["pinned"]

    result = run(cli.statuses.unpin, status["id"])
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Status unpinned"

    status = api.fetch_status(app, user, status["id"]).json()
    assert not status["pinned"]


def test_pin_json(app, user, run):
    status = api.post_status(app, user, "foo").json()
    assert not status["pinned"]

    result = run(cli.statuses.pin, status["id"], "--json")
    assert result.exit_code == 0

    result = json.loads(result.stdout)
    assert result["pinned"] is True
    assert result["id"] == status["id"]

    result = run(cli.statuses.unpin, status["id"], "--json")
    assert result.exit_code == 0

    result = json.loads(result.stdout)
    assert result["pinned"] is False
    assert result["id"] == status["id"]


def test_bookmark(app, user, run):
    status = api.post_status(app, user, "foo").json()
    assert not status["bookmarked"]

    result = run(cli.statuses.bookmark, status["id"])
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Status bookmarked"

    status = api.fetch_status(app, user, status["id"]).json()
    assert status["bookmarked"]

    result = run(cli.statuses.unbookmark, status["id"])
    assert result.exit_code == 0
    assert result.stdout.strip() == "✓ Status unbookmarked"

    status = api.fetch_status(app, user, status["id"]).json()
    assert not status["bookmarked"]


def test_bookmark_json(app, user, run):
    status = api.post_status(app, user, "foo").json()
    assert not status["bookmarked"]

    result = run(cli.statuses.bookmark, status["id"], "--json")
    assert result.exit_code == 0

    result = json.loads(result.stdout)
    assert result["id"] == status["id"]
    assert result["bookmarked"] is True

    result = run(cli.statuses.unbookmark, status["id"], "--json")
    assert result.exit_code == 0

    result = json.loads(result.stdout)
    assert result["id"] == status["id"]
    assert result["bookmarked"] is False



import toot
from pkg_resources import get_distribution


def test_version():
    """Version specified in __version__ should be the same as the one
    specified in setup.py."""
    assert toot.__version__ == get_distribution('toot').version



import urwid

from toot.tui.utils import highlight_hashtags
from toot.utils import format_content
from typing import List

try:
    from .richtext import html_to_widgets, url_to_widget
except ImportError:
    # Fallback if urwidgets are not available
    def html_to_widgets(html: str) -> List[urwid.Widget]:
        return [
            urwid.Text(highlight_hashtags(line))
            for line in format_content(html)
        ]

    def url_to_widget(url: str):
        return urwid.Text(("link", url))



import json

from toot import App, User, api, cli
from toot.entities import Account, Relationship, from_dict


def test_whoami(user: User, run):
    result = run(cli.read.whoami)
    assert result.exit_code == 0

    # TODO: test other fields once updating account is supported
    out = result.stdout.strip()
    assert f"@{user.username}" in out


def test_whoami_json(user: User, run):
    result = run(cli.read.whoami, "--json")
    assert result.exit_code == 0

    account = from_dict(Account, json.loads(result.stdout))
    assert account.username == user.username


def test_whois(app: App, friend: User, run):
    variants = [
        friend.username,
        f"@{friend.username}",
        f"{friend.username}@{app.instance}",
        f"@{friend.username}@{app.instance}",
    ]

    for username in variants:
        result = run(cli.read.whois, username)
        assert result.exit_code == 0
        assert f"@{friend.username}" in result.stdout


def test_following(app: App, user: User, friend: User, friend_id, run):
    # Make sure we're not initally following friend
    api.unfollow(app, user, friend_id)

    result = run(cli.accounts.following, user.username)
    assert result.exit_code == 0
    assert result.stdout.strip() == ""

    result = run(cli.accounts.follow, friend.username)
    assert result.exit_code == 0
    assert result.stdout.strip() == f"✓ You are now following {friend.username}"

    result = run(cli.accounts.following, user.username)
    assert result.exit_code == 0
    assert friend.username in result.stdout.strip()

    # If no account is given defaults to logged in user
    result = run(cli.accounts.following)
    assert result.exit_code == 0
    assert friend.username in result.stdout.strip()

    result = run(cli.accounts.unfollow, friend.username)
    assert result.exit_code == 0
    assert result.stdout.strip() == f"✓ You are no longer following {friend.username}"

    result = run(cli.accounts.following, user.username)
    assert result.exit_code == 0
    assert result.stdout.strip() == ""


def test_following_case_insensitive(user: User, friend: User, run):
    assert friend.username != friend.username.upper()
    result = run(cli.accounts.follow, friend.username.upper())
    assert result.exit_code == 0

    out = result.stdout.strip()
    assert out == f"✓ You are now following {friend.username.upper()}"


def test_following_not_found(run):
    result = run(cli.accounts.follow, "bananaman")
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: Account not found"

    result = run(cli.accounts.unfollow, "bananaman")
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: Account not found"


def test_following_json(app: App, user: User, friend: User, user_id, friend_id, run_json):
    # Make sure we're not initally following friend
    api.unfollow(app, user, friend_id)

    result = run_json(cli.accounts.following, user.username, "--json")
    assert result == []

    result = run_json(cli.accounts.followers, friend.username, "--json")
    assert result == []

    result = run_json(cli.accounts.follow, friend.username, "--json")
    relationship = from_dict(Relationship, result)
    assert relationship.id == friend_id
    assert relationship.following is True

    [result] = run_json(cli.accounts.following, user.username, "--json")
    relationship = from_dict(Relationship, result)
    assert relationship.id == friend_id

    # If no account is given defaults to logged in user
    [result] = run_json(cli.accounts.following, user.username, "--json")
    relationship = from_dict(Relationship, result)
    assert relationship.id == friend_id

    [result] = run_json(cli.accounts.followers, friend.username, "--json")
    assert result["id"] == user_id

    result = run_json(cli.accounts.unfollow, friend.username, "--json")
    assert result["id"] == friend_id
    assert result["following"] is False

    result = run_json(cli.accounts.following, user.username, "--json")
    assert result == []

    result = run_json(cli.accounts.followers, friend.username, "--json")
    assert result == []


def test_mute(app, user, friend, friend_id, run):
    # Make sure we're not initially muting friend
    api.unmute(app, user, friend_id)

    result = run(cli.accounts.muted)
    assert result.exit_code == 0

    out = result.stdout.strip()
    assert out == "No accounts muted"

    result = run(cli.accounts.mute, friend.username)
    assert result.exit_code == 0

    out = result.stdout.strip()
    assert out == f"✓ You have muted {friend.username}"

    result = run(cli.accounts.muted)
    assert result.exit_code == 0

    out = result.stdout.strip()
    assert friend.username in out

    result = run(cli.accounts.unmute, friend.username)
    assert result.exit_code == 0

    out = result.stdout.strip()
    assert out == f"✓ {friend.username} is no longer muted"

    result = run(cli.accounts.muted)
    assert result.exit_code == 0

    out = result.stdout.strip()
    assert out == "No accounts muted"


def test_mute_case_insensitive(friend: User, run):
    result = run(cli.accounts.mute, friend.username.upper())
    assert result.exit_code == 0

    out = result.stdout.strip()
    assert out == f"✓ You have muted {friend.username.upper()}"


def test_mute_not_found(run):
    result = run(cli.accounts.mute, "doesnotexistperson")
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: Account not found"

    result = run(cli.accounts.unmute, "doesnotexistperson")
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: Account not found"


def test_mute_json(app: App, user: User, friend: User, run_json, friend_id):
    # Make sure we're not initially muting friend
    api.unmute(app, user, friend_id)

    result = run_json(cli.accounts.muted, "--json")
    assert result == []

    result = run_json(cli.accounts.mute, friend.username, "--json")
    relationship = from_dict(Relationship, result)
    assert relationship.id == friend_id
    assert relationship.muting is True

    [result] = run_json(cli.accounts.muted, "--json")
    account = from_dict(Account, result)
    assert account.id == friend_id

    result = run_json(cli.accounts.unmute, friend.username, "--json")
    relationship = from_dict(Relationship, result)
    assert relationship.id == friend_id
    assert relationship.muting is False

    result = run_json(cli.accounts.muted, "--json")
    assert result == []


def test_block(app, user, friend, friend_id, run):
    # Make sure we're not initially blocking friend
    api.unblock(app, user, friend_id)

    result = run(cli.accounts.blocked)
    assert result.exit_code == 0

    out = result.stdout.strip()
    assert out == "No accounts blocked"

    result = run(cli.accounts.block, friend.username)
    assert result.exit_code == 0

    out = result.stdout.strip()
    assert out == f"✓ You are now blocking {friend.username}"

    result = run(cli.accounts.blocked)
    assert result.exit_code == 0

    out = result.stdout.strip()
    assert friend.username in out

    result = run(cli.accounts.unblock, friend.username)
    assert result.exit_code == 0

    out = result.stdout.strip()
    assert out == f"✓ {friend.username} is no longer blocked"

    result = run(cli.accounts.blocked)
    assert result.exit_code == 0

    out = result.stdout.strip()
    assert out == "No accounts blocked"


def test_block_case_insensitive(friend: User, run):
    result = run(cli.accounts.block, friend.username.upper())
    assert result.exit_code == 0

    out = result.stdout.strip()
    assert out == f"✓ You are now blocking {friend.username.upper()}"


def test_block_not_found(run):
    result = run(cli.accounts.block, "doesnotexistperson")
    assert result.exit_code == 1
    assert result.stderr.strip() == "Error: Account not found"


def test_block_json(app: App, user: User, friend: User, run_json, friend_id):
    # Make sure we're not initially blocking friend
    api.unblock(app, user, friend_id)

    result = run_json(cli.accounts.blocked, "--json")
    assert result == []

    result = run_json(cli.accounts.block, friend.username, "--json")
    relationship = from_dict(Relationship, result)
    assert relationship.id == friend_id
    assert relationship.blocking is True

    [result] = run_json(cli.accounts.blocked, "--json")
    account = from_dict(Account, result)
    assert account.id == friend_id

    result = run_json(cli.accounts.unblock, friend.username, "--json")
    relationship = from_dict(Relationship, result)
    assert relationship.id == friend_id
    assert relationship.blocking is False

    result = run_json(cli.accounts.blocked, "--json")
    assert result == []



import json
import os

from contextlib import contextmanager
from os.path import dirname, join
from typing import Optional

from toot import User, App, get_config_dir
from toot.exceptions import ConsoleError


TOOT_CONFIG_FILE_NAME = "config.json"


def get_config_file_path():
    """Returns the path to toot config file."""
    return join(get_config_dir(), TOOT_CONFIG_FILE_NAME)


def user_id(user):
    return "{}@{}".format(user.username, user.instance)


def make_config(path):
    """Creates an empty toot configuration file."""
    config = {
        "apps": {},
        "users": {},
        "active_user": None,
    }

    # Ensure dir exists
    os.makedirs(dirname(path), exist_ok=True)

    # Create file with 600 permissions since it contains secrets
    fd = os.open(path, os.O_CREAT | os.O_WRONLY, 0o600)
    with os.fdopen(fd, 'w') as f:
        json.dump(config, f, indent=True)


def load_config():
    # Just to prevent accidentally running tests on production
    if os.environ.get("TOOT_TESTING"):
        raise Exception("Tests should not access the config file!")

    path = get_config_file_path()

    if not os.path.exists(path):
        make_config(path)

    with open(path) as f:
        return json.load(f)


def save_config(config):
    path = get_config_file_path()
    with open(path, "w") as f:
        return json.dump(config, f, indent=True, sort_keys=True)


def extract_user_app(config, user_id):
    if user_id not in config['users']:
        return None, None

    user_data = config['users'][user_id]
    instance = user_data['instance']

    if instance not in config['apps']:
        return None, None

    app_data = config['apps'][instance]
    return User(**user_data), App(**app_data)


def get_active_user_app():
    """Returns (User, App) of active user or (None, None) if no user is active."""
    config = load_config()

    if config['active_user']:
        return extract_user_app(config, config['active_user'])

    return None, None


def get_user_app(user_id):
    """Returns (User, App) for given user ID or (None, None) if user is not logged in."""
    return extract_user_app(load_config(), user_id)


def load_app(instance: str) -> Optional[App]:
    config = load_config()
    if instance in config['apps']:
        return App(**config['apps'][instance])


def load_user(user_id, throw=False):
    config = load_config()

    if user_id in config['users']:
        return User(**config['users'][user_id])

    if throw:
        raise ConsoleError("User '{}' not found".format(user_id))


def get_user_list():
    config = load_config()
    return config['users']


@contextmanager
def edit_config():
    config = load_config()
    yield config
    save_config(config)


def save_app(app: App):
    with edit_config() as config:
        config['apps'][app.instance] = app._asdict()


def delete_app(config, app):
    with edit_config() as config:
        config['apps'].pop(app.instance, None)


def save_user(user: User, activate=True):
    with edit_config() as config:
        config['users'][user_id(user)] = user._asdict()

        if activate:
            config['active_user'] = user_id(user)


def delete_user(user: User):
    with edit_config() as config:
        config['users'].pop(user_id(user), None)

        if config['active_user'] == user_id(user):
            config['active_user'] = None


def activate_user(user: User):
    with edit_config() as config:
        config['active_user'] = user_id(user)
